"""
Stores the Configuration for a given simulation: habitat heat loading and external environment
"""

from syrtis.material import *


solution_types = ["constant power", "constant temperature"]
air_directions = ["axial", "cross"]

class Configuration:
    def __init__(self, name, solution_type,
    T_ground, k_ground, albedo_ground, T_air, p_air, v_air, air_direction, solar_altitude, solar_azimuth, solar_intensity,
    Q_habitat=0, T_habitat=0):
        """
        An object to store the Configuration for a given simulation

        Args:
            name (str):             an identifier for the configuration
            solution_type(str):     either "constant power" or "constant temperature" for now, more may be added
            
            T_ground (float):       ground/soil temperature (K)
            k_ground (float):       soil thermal conductivity (W/m/K)
            albedo_ground (float)   albedo of the ground, between 0 and 1
            
            T_air (float):          air temperature (K)
            p_air (float):          air pressure (Pa)
            v_air (float):          speed of the air (m/s)
            air_direction (float):  direction of the air, either "axial" for along a horizontal cyclinder or "cross" for crossflow

            solar_altitude (float): vertical angle of the sun above the horizon (degrees)
            solar_azimuth (float):  horizontal angle of the sun RELATIVE TO HABITAT AXIS (degrees) - 0=directly along axis
            solar_intensity (float):power delivered by solar radiation after dust absorption, W/m2

            Q_habitat (float):      constant power generation in the Habitat (W) - optional, defaults to zero
            T_habitat (float):      constant temeprature in the Habitat (K) - optional unless solution_type is "constant temperature"  
        """

        assert solution_type in solution_types, "Configuration 'solution_type' must be a valid keyword"

        #assert solution_type == "constant temperature" and T_habitat != 0, "If 'constant temperature' has been selected, T_habitat must be set"

        assert air_direction in air_directions, "Configuration 'air_direction' must be a valid keyword"

        assert is_numeric(albedo_ground, unit=True), "Configuration 'albedo_ground' must be between 0 and 1"

        self.name = name
        self.solution_type = solution_type

        self.T_ground = T_ground
        self.k_ground = k_ground
        self.albedo_ground = albedo_ground

        self.T_air = T_air
        self.p_air = p_air
        self.v_air = v_air
        self.air_direction = air_direction

        self.solar_altitude = solar_altitude
        self.solar_azimuth = solar_azimuth
        self.solar_intensity = solar_intensity

        self.Q_habitat = Q_habitat
        self.T_habitat = T_habitat

        self.GRAVITY = 3.71
        # Assume we're only using Mars for now

        self.air = ConstrainedIdealGas("Surface air", self.p_air, 44, 0.71, 10.9e-6, 749, 0.0153)


"""
Miscellaneous tool functions for use elsewhere

"""

from numbers import Number
import numpy as np
import matplotlib.pyplot as plt

def is_numeric(x, positive=False, not_negative=False, unit=False):
    """
    Check if input x is either numeric, or is a list of numeric
    """
    if isinstance(x, Number):
        if positive:
            # x must be greater than zero
            if x > 0:
                return(True)
            else:
                return(False)

        elif not_negative:
            # x must be greater than or equal to zero
            if x >= 0:
                return(True)
            else:
                return(False)

        elif unit:
            # x must be between 0 and 1
            if x>=0 and x<=1:
                return(True)
            else:
                return(False)

        else:
            return(True)

    elif type(x) == tuple or type(x) == list:
        all_items_numeric = all(is_numeric(item, positive, not_negative) for item in x)
        return(all_items_numeric)
    
    else:
        return(False)

def plot_power_balance(heat_reports, labels):
    """"
    
        heat_reports (list of dicts):    list of standardised reports
    """
    
    gain_keys = []
    gain_values = [ [] for _ in range(len(heat_reports)) ]
    loss_keys = []
    loss_values = [ [] for _ in range(len(heat_reports)) ]

    for key in heat_reports[0].keys():
        if "gain" in key:
            gain_keys.append(key)
            for index, report in enumerate(heat_reports):
                gain_values[index].append(report[key])
        
        elif "loss" in key:
            loss_keys.append(key)
            for index, report in enumerate(heat_reports):
                loss_values[index].append(report[key])
    
    
    for index in range(len(heat_reports)):
        gain_values.insert((2*index)+1, 
            np.zeros(len(gain_values[0])))
        loss_values.insert((2*index), 
            np.zeros(len(loss_values[0])))

    cum_gain = np.zeros(len(heat_reports)*2)
    gain_values = np.asarray(gain_values)
    loss_values = np.asarray(loss_values)
    all_keys = gain_keys + loss_keys

    #labels = ["a", "b"]

    x = np.arange(len(labels))  # the label locations
    width = 0.35  # the width of the bars

    shifted=False
    
    for value_set in range(len(all_keys)):
        if all_keys[value_set] in gain_keys:
            plt.bar(x-width/2, 
            gain_values[::2,value_set], 
            width, 
            label=all_keys[value_set], 
            bottom=cum_gain[::2])


            for x_item in range(len(x)):
                if gain_values[x_item*2,value_set] != 0:
                    plt.text(x[x_item]-width, 
                        gain_values[x_item*2,value_set]/2 + cum_gain[x_item*2], 
                        all_keys[value_set], 
                        va="center")

            cum_gain = np.add(cum_gain, gain_values[:,value_set])

            

        elif all_keys[value_set] in loss_keys:

            offset_value_set = value_set - len(gain_keys)

            plt.bar(x+width/2, 
            loss_values[1::2,offset_value_set], 
            width, 
            label=all_keys[value_set], 
            bottom=np.add(cum_gain[1::2], cum_gain[::2]))

            for x_item in range(len(x)):
                if loss_values[1 + x_item*2,offset_value_set] != 0:
                    plt.text(x[x_item], 
                        loss_values[1 + x_item*2,offset_value_set]/2 + np.add(cum_gain[1 + x_item*2], cum_gain[x_item*2]), 
                        all_keys[value_set], 
                        va="center")
            
            cum_gain = np.add(cum_gain, loss_values[:,offset_value_set])

        
    plt.xticks(x, labels) 
    plt.axhline(0, color='black')

    plt.ylabel("Heat loss into habitat (W) - positive=loss")   
    
    plt.legend(bbox_to_anchor = (1.05, 0.6))
    
    
    


import numpy as np
from syrtis.tools import *

class Material:
    """
    Placeholder
    """

class Solid(Material):
    """
    Object for a homogenous solid Shell

    Args:
        name (str):         Name of the material
        k (float):          Thermal conductivity (W/m/K)
        rho (float):        Density (kg/m3)
        cp (float):         Specific heat capacity (kJ/kg/K)
        absorb (float):     Absorptivity, between 0 and 1
        emit (float):       Emissivity, between 0 and 1. Optional, defaults equal to absorbivity
        transmit (float):   Transmissivity, between 0 and 1, Optional, defaults to zero
    """
    def __init__(self, name, k, rho, cp, absorb, emit=0,  transmit=0):
        assert is_numeric(k, positive=True),   "Material 'k' must be a positive numerical value"
        assert is_numeric(rho, positive=True),   "Material 'rho' must be a positive numerical value"
        assert is_numeric(cp, positive=True), "Material 'cp' must be a positive numerical value"
        assert is_numeric(absorb, unit=True), "Material 'absorb' must be between 0 and 1"
        assert is_numeric(emit, unit=True), "Material 'emit' must be between 0 and 1"
        assert is_numeric(transmit, unit=True), "Material 'alpha' must be between 0 and 1"

        self.name = name
        self.k = k
        self.rho = rho
        self.cp = cp
        self.absorb = absorb

        if emit == 0:
            self.emit = self.absorb
        else:
            self.emit = emit
            
        self.transmit = transmit

class ConstrainedIdealGas(Material):
    """
    Object for a constrained ideal gas with no imposed flow (undergoing natural convection) and at constant pressure
    All temperature-dependent properties (density, beta) are calculated in functions of the same name
    Reference specific heat capacity cp_ref is given at standard temperature and pressure (101.3kPa, 298K)

    Args:
        p (float):      Pressure (Pa)
        M (float):      Molecular mass number
        Pr (float):     Prandtl number
        mu (float):     Absolute viscosity (Pa s)
        cp_ref (float): Reference isobaric specific heat capacity (kJ/kg/K)
        k (float):      Thermal conductivity (W/m/K)
        p (float):      Pressure (Pa) - optional, used as constant pressure if filled
        T (float):      Temperature (K) - optional, used as constant temperature if filled
        beta (float):   Thermal expansion coefficient (1/K) - optional, set to 1/T if left blank
    """
    def __init__(self, name, p, M, Pr, mu, cp, k, T=0, beta=0):
        assert is_numeric(p, positive=True),     "Material 'p' must be a positive numerical value"
        assert is_numeric(M, positive=True),     "Material 'M' must be a positive numerical value"
        assert is_numeric(Pr, positive=True),   "Material 'Pr' must be a positive numerical value"
        assert is_numeric(mu, positive=True),   "Material 'mu' must be a positive numerical value"
        assert is_numeric(cp, positive=True),   "Material 'cp' must be a positive numerical value"
        assert is_numeric(k, positive=True),     "Material 'k' must be a positive numerical value"

        if beta != 0:
            assert is_numeric(beta, positive=True),   "Material 'beta' must be a positive numerical value"

        if T != 0:
            assert is_numeric(T, positive=True),   "Material 'T' must be a positive numerical value"

        self.name = name
        self._p = p
        self._M = M
        self._Pr = Pr
        self._mu = mu
        self._cp = cp
        self._k = k

        if beta == 0:
            self.input_beta = True
            self._beta = 0.0
        else:
            self.input_beta = False
            self._beta = beta
        
        if T == 0:
            self.input_T = True
            self._T = 0.0
        else:
            self.input_T = False
            self._T = T
    
    def rho(self, T=0):
        """
        Gas density. If T is held constant, the input values will be ignored

        Args:
            T (float):  Temperature (K) 
        """
        
        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"
        
        rho = self._p / ((8314 / self._M) * T) 

        return(rho)
    
    def beta(self, T):
        """
        Coefficient of thermal expansion, equal to 1/T. If T is held constant, the input value is ignored
        
        Args:
            T (float):  Temperature (K) 
        """

        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"

        beta = 1/T

        return(beta)

    def mu(self, T):
        """
        Dynamic viscosity

        Args:
            T (float):  Temperature (K) 
        """

        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"

        return(self._mu)

    def cp(self, T):
        """
        Specific heat capacity at constant pressure

        Args:
            T (float):  Temperature (K) 
        """

        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"

        return(self._cp)

    def Pr(self, T):
        """
        Prandtl number

        Args:
            T (float):  Temperature (K) 
        """

        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"

        return(self._Pr)
    
    def k(self, T):
        """
        Thermal conductivity

        Args:
            T (float):  Temperature (K) 
        """

        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"

        return(self._k)
    
    def Re(self, T, L, v):
        """
        Dynamic viscosity

        Args:
            T (float):  Temperature (K) 
        """

        if self.input_T == False:
            T = self._T
        else:
            assert is_numeric(T, positive=True),   "Input 'T' must be a positive numerical value"
        
        assert is_numeric(L, positive=True),   "Input 'L' must be a positive numerical value"
        assert is_numeric(v, positive=True),   "Input 'v' must be a positive numerical value"

        Re = self.rho(T) * v * L / self.mu(T)

        return(Re)

    def Ra(self, T_avg, T_delta, g, length):
        """
        Calculate the Rayleigh number, using constant fluid properties

        
        """
        
        if self.input_T == False:
            T_avg = self._T
        else:
            assert is_numeric(T_avg, positive=True),    "Input 'T_avg' must be a positive numerical value"

        assert is_numeric(T_delta),      "Input 'T_delta' must be a numerial value"

        assert is_numeric(g, positive=True),            "Input 'g' must be a positive numerical value"
        assert is_numeric(length, positive=True),       "Input 'length' must be a positive numerical value"

        Ra = ((g * np.power(length, 3) * self.beta(T_avg) * abs(T_delta) * np.power(self.rho(T_avg), 2)) 
        / np.power(self.mu(T_avg), 2))

        assert Ra > 0, "Input has produced an invalid output"

        return(Ra)




material_classes = [Solid, ConstrainedIdealGas]

ambient_atmosphere = ConstrainedIdealGas("Martian ambient CO2", 580, 44, 0.71, 10.9e-6, 749, 0.0153, T=210)



"""
Object for sweeping through different parameters in the Solver

References:
 - [1] - A solar azimuth formula that renders circumstantial treatment unnecessary without compromising mathematical rigor: 
        Mathematical setup, application and extension of a formula based on the subsolar point and atan2 function, Zhang et al 2021
 - [2] - Über die Extinktion des Lichtes in der Erdatmosphäre, Schoenberg 1929
 - [3] - Mars Fact Sheet, https://nssdc.gsfc.nasa.gov/planetary/factsheet/marsfact.html
 - [4] - Thermal control of MSL Rover "Curiosity" using an Active Fluid Loop, Birur 2013 
"""
import numpy as np
import copy

from syrtis.configuration import Configuration
from syrtis.solver import Solver
from itertools import product

SOL_HRS = 24.65
MARS_RADIUS = 3396.2e3
MARS_ATMOSPHERE_OPTICAL_HEIGHT = 8624
MARS_ECC = 0.093377


class SolverManager:
    """
    Parent class for all other solver managers
    """


    def __init__(self):
        pass

    def run_all_configurations(self, verbose=False):
        """
        Run all the configurations
        """

        heat_losses = []
        reports = []
        completed_solvers = []
        
        for configuration_index, configuration in enumerate(self.configurations):
            self.habitat.verify_geometry()

            solver = Solver(str(self.each_configuration_inputs_dicts[configuration_index]["name"]),
                self.habitat, configuration)
            
            solver.solve()
            heat_losses.append(solver.heat)
            reports.append(solver.report)
            completed_solvers.append(solver)
            
        self.heat_losses = heat_losses
        self.reports = reports
        self.completed_solvers = completed_solvers

        if verbose:
            return(self.each_configuration_inputs_dicts, self.heat_losses, self.reports)
        else:
            return(self.each_configuration_inputs_dicts, self.heat_losses)

class ConfigurationManager(SolverManager):
    """
    Sweep through different parameters in the Configuration
    """
    configuration_variables = ["name", "solution_type",
        "T_ground", "k_ground", "albedo_ground", "T_air", "p_air", "v_air", "air_direction", "solar_altitude", "solar_azimuth", "solar_intensity",
        "Q_habitat", "T_habitat"]

    def __init__(self, habitat, configuration, variable_parameters, all_configurations=[]):
        
        assert all(parameter in self.configuration_variables for parameter in list(variable_parameters.keys())), "Not all configuration parameters set"

        self.habitat = habitat
        self.configuration_baseline = configuration

        self.variable_parameters = variable_parameters

        if all_configurations == []:
            self.configurations = self.create_configurations()
        else:
            self.configurations = all_configurations

    def create_configurations(self):
          
        iterations_number = []
        inputs_keys = []
        iterations_values = []

        for key in self.configuration_variables:
            if not key in self.variable_parameters:
                pass
            elif type(self.variable_parameters[key]) == list or type(self.variable_parameters[key]) == tuple:
                iterations_number.append(len(self.variable_parameters[key]))
                inputs_keys.append(key)
                iterations_values.append(self.variable_parameters[key])
        
        self.all_configuration_inputs = np.zeros(iterations_number)

        each_configuration_inputs = list(product(*iterations_values))


        self.each_configuration_inputs_dicts = []
        
        for config_input in each_configuration_inputs:
            input_dict = copy.deepcopy(self.configuration_baseline.__dict__)

            for input_index, input in enumerate(config_input):
                input_dict[inputs_keys[input_index]] = input

            self.each_configuration_inputs_dicts.append(input_dict)

        all_configurations = []

        for input_dict in self.each_configuration_inputs_dicts:
            config = Configuration(
                name = input_dict["name"], 
                solution_type = input_dict["solution_type"],
                T_ground = input_dict["T_ground"], 
                k_ground = input_dict["k_ground"], 
                albedo_ground = input_dict["albedo_ground"], 
                T_air = input_dict["T_air"], 
                p_air = input_dict["p_air"], 
                v_air = input_dict["v_air"], 
                air_direction = input_dict["air_direction"], 
                solar_altitude = input_dict["solar_altitude"], 
                solar_azimuth = input_dict["solar_azimuth"],
                solar_intensity = input_dict["solar_intensity"],
                Q_habitat = input_dict["Q_habitat"], 
                T_habitat = input_dict["T_habitat"]
            )
            all_configurations.append(config)
        
        return(all_configurations)

class DayManager(SolverManager):
    """
    Calculate heat flux at each point in a day
    
    Args:
        habitat (syrtis.Habitat):               habitat under consideration
        configuration (syrtis.Configuration):   baseline configuration - temperatures and solar properties will be overwritten

    """
    def __init__(self, habitat, configuration, num_points,
        atmosphere_tau, latitude, areocentric_longitude,
        T_air_max, T_air_min, T_ground_max, T_ground_min, time_air_peak=12, time_ground_peak=12):

        self.habitat = habitat
        self.configuration = configuration
        self.num_points = num_points

        self.atmosphere_tau = atmosphere_tau
        self.latitude = latitude
        self.areocentric_longitude = areocentric_longitude

        self.T_air_max = T_air_max
        self.T_air_min = T_air_min
        self.T_ground_max = T_ground_max
        self.T_ground_min = T_ground_min

        self.time_air_peak = time_air_peak
        self.time_ground_peak = time_ground_peak

        self.create_configurations()
              
    def create_configurations(self):
        """
        Create all the configurations for the day
        """
        self.times = np.linspace(0, SOL_HRS, self.num_points)
        self.generate_solar_data()
        self.T_airs = self.generate_temperatures(self.T_air_max, self.T_air_min, self.time_air_peak)
        self.T_grounds = self.generate_temperatures(self.T_ground_max, self.T_ground_min, self.time_ground_peak)

        baseline_configuration = copy.deepcopy(self.configuration)

        configurations = []
        each_configuration_inputs_dicts = []
        
        for point, time in enumerate(self.times):

            time_configuration = copy.deepcopy(baseline_configuration)
            time_configuration.name = time_configuration.name + " {:.2f}".format(time)

            time_configuration.T_air = self.T_airs[point]
            time_configuration.T_ground = self.T_grounds[point]

            time_configuration.solar_intensity = self.solar_intensities[point]
            time_configuration.solar_altitude = self.solar_altitudes[point]
            time_configuration.solar_azimuth = self.solar_azimuths[point]

            configuration_dict = copy.deepcopy(time_configuration.__dict__)

            configurations.append(time_configuration)
            each_configuration_inputs_dicts.append(configuration_dict)
        
        self.configurations = configurations
        self.each_configuration_inputs_dicts = each_configuration_inputs_dicts
    
    def generate_solar_data(self):
        """
        Generate lists of solar intensities, altitudes and azimuths for a single day.

        Solar position implements method from Reference [1] with necessary modifications for Martian orbital parameters.

        Martian atmospheric optimal height, for use in calculation from Source [2]. Calculation is based on constant-density
        hydrostatic argument, with data taken from [3]
        """
        
        axis_obliquity = np.deg2rad(24.936)
        subsolar_latitude = np.arcsin(np.sin(axis_obliquity) * np.sin(np.deg2rad(self.areocentric_longitude)))

        solar_intensity_space = 590 * np.power(
        (1 + MARS_ECC * np.cos(np.deg2rad(self.areocentric_longitude - 248))) / (1 - MARS_ECC), 2)

        solar_intensities = np.zeros((self.num_points))
        solar_altitudes = np.zeros((self.num_points))
        solar_azimuths = np.zeros((self.num_points))


        for point in range(self.num_points):
            subsolar_longitude = np.deg2rad(-15 * (self.times[point] - (SOL_HRS/2)))

            Sx = np.cos(subsolar_latitude) * np.sin(subsolar_longitude)
            Sy = (np.cos(self.latitude) * np.sin(subsolar_latitude)) - (
                np.sin(self.latitude) * np.cos(subsolar_latitude) * np.cos(subsolar_longitude))
            Sz = (np.sin(np.deg2rad(self.latitude)) * np.sin(subsolar_latitude)) + (
                np.cos(np.deg2rad(self.latitude)) * np.cos(subsolar_latitude) * np.cos(subsolar_longitude))

            solar_altitude = np.rad2deg(np.arcsin(Sz))
            solar_azimuth_south_clockwise = np.rad2deg(np.arctan2(-Sx, Sy))
            solar_azimuth = solar_azimuth_south_clockwise + self.configuration.solar_azimuth


            # Calculate relative air mass for scaling the optical depth of the atmosphere
            relative_air_mass = np.sqrt(np.power(MARS_RADIUS + MARS_ATMOSPHERE_OPTICAL_HEIGHT, 2) - 
                np.power(MARS_RADIUS * np.cos(np.deg2rad(solar_altitude)), 2)) - MARS_RADIUS * np.sin(np.deg2rad(solar_altitude))
            relative_air_mass /= MARS_ATMOSPHERE_OPTICAL_HEIGHT

            optical_depth = np.exp(- self.atmosphere_tau * relative_air_mass)

            solar_intensity = solar_intensity_space * optical_depth

            if Sz <= 0:
                solar_intensity = 0
                solar_altitude = 0

            solar_intensities[point] = solar_intensity
            solar_altitudes[point] = solar_altitude
            solar_azimuths[point] = solar_azimuth
        
        self.solar_intensities = solar_intensities
        self.solar_altitudes = solar_altitudes
        self.solar_azimuths = solar_azimuths

    def generate_temperatures(self, T_peak, T_min, time_peak):
        """
        Generate the temperatures at each point in the day - can be called with either ground or ambient

        Model used is a sinusoidal fit during the day, rising from a minimum at sunset to a maximum
            either at local noon, or time_peak if set
        Between sunset and sunrise, a linear fit is used to bridge the temperatures.
        This approximately matches the data from [4]
        """
        
        axis_obliquity = np.deg2rad(24.936)
        solar_declination = np.arcsin(np.sin(axis_obliquity) * np.sin(np.deg2rad(self.areocentric_longitude)))

        time_sunrise = (SOL_HRS / (2 * np.pi)) * np.arccos(-np.tan(np.deg2rad(self.latitude)) * np.tan(solar_declination))
        time_sunset = SOL_HRS - time_sunrise

        # Setting up amplitude for the cos wave
        T_sine_peak = T_peak 
        T_sine_trough = (T_min - T_sine_peak * np.cos((time_sunrise - time_peak) * 2 * np.pi / SOL_HRS)) / (
            1 - np.cos((time_sunrise - time_peak) * 2 * np.pi / SOL_HRS))
        
        def T_sinusoid(time):
            T = (T_sine_peak - T_sine_trough) * (np.cos((time - time_peak) * 2 * np.pi / SOL_HRS)) + T_sine_trough
            return(T)

        # Set up parameters for the linear fit
        T_sunset = T_sinusoid(time_sunset)
        T_sunrise = T_min
        T_midnight = (T_sunset + T_sunrise) / 2

        def T_linear(time):
            if time > time_sunset:
                T = (T_sunset - T_midnight) * ((SOL_HRS - time) / (SOL_HRS - time_sunset)) + T_midnight
            else:
                T = (T_sunrise - T_midnight) * (time / time_sunrise) + T_midnight
            return(T)
        
        temps = np.zeros((self.num_points))


        for point, time in enumerate(self.times):
            if time > time_sunrise and time < time_sunset:
                # Time is during the day
                T = T_sinusoid(time)
            else:
                T = T_linear(time)
            
            temps[point] = T
        
        return(temps)

        








from syrtis.shell import *
from syrtis.material import *
from syrtis.habitat import *
from syrtis.configuration import *
from syrtis.solver import *
from syrtis.solvermanager import *
from syrtis.tools import *

import syrtis.shell
import syrtis.material
import syrtis.habitat


"""
Top-level object and tools for the whole Habitat, composed of individual Shells

References:

 - [1] - Y Cengel, Heat Transfer 2nd edition
 - [2] - Sky temperature modelisation and applications in building simulation, Adelard et al 1998
 - [3] - Long-wave  radiation from clear skies, Swinbank 1963
 - [4] - Radiation View Factors, http://webserver.dmt.upm.es/~isidoro/tc3/Radiation%20View%20factors.pdf
 - [5] - Thermal control of MSL Rover "Curiosity" using an Active Fluid Loop, Birur 2013
 - [6] - The thermal control system of NASA’s Curiosity rover: a case study, Quattrocchi et al, 2022 
 - [7] - Conduction heat transfer solutions, VanSant 1980
 - [8] - Thermal Spreading Resistance of Arbitrary-Shape Heat Sources on a Half-Space: A Unified Approach, Sadeghi et al 2010
 - [9] - Heat transfer from partially buried pipes, Morud & Simonsen 2007
 - [10]- Proposed OHTC Formula for Subsea Pipelines Considering Thermal Conductivities of Multi-Layered Soils, Park et al 2018
 - [11]- Heat Transfer, J P Holman 10th Edition
"""

import numpy as np
from random import choice
import matplotlib.patches as patches


if __name__ == "__main__":
    from shell import *
    from material import *
else:
    #from shell import *
    from syrtis.shell import *
    from syrtis.material import *

endcap_types = ["hemisphere", "flat"]

class Habitat:
    """
    Object to store the whole habitat geometry

    Args:
        orientation (str):        either 'horizontal' or 'vertical' to specify orientation of axis
        length (float):             the length of the central axis of the Habitat (m)    
    """

    def __init__(self, orientation, length, endcap_type):
        assert orientation == "horizontal" or orientation == "vertical", "'orientation' must be either 'horizontal' or 'vertical'"

        assert is_numeric(length, not_negative=True), "Habitat 'length' must be a positive numerical value"

        assert endcap_type in endcap_types, "'endcap_type' must be a valid keyword"

        self.orientation = orientation
        self.length = length

        self.endcap_type = endcap_type

        self._shells = []
        self.groundlevel = None
        self.earthworks = None

        self.radius_outer = 0
        self.length_outer = 0
        self.ground_contact_angle = 0

        self.verified = False
    
    def draw(self,size=(20,20)):
        """
        Sets up a matplotlib object to draw the Habitat
        """
        self.verify_geometry()

        fig = plt.figure(figsize=size)

        if self.orientation == "horizontal":
            max_x = self.radius_outer * 1.25
            max_y = self.radius_outer * 1.25
            if self.endcap_type == "hemisphere":
                max_z = (self.length / 2) + (self.radius_outer * 1.25)
            else:
                max_z = (self.length / 2) * 1.25
        elif self.orientation == "vertical":
            max_x = self.radius_outer * 1.25
            if self.endcap_type == "hemisphere":
                max_y = (self.length / 2) + (self.radius_outer * 1.25)
            else:
                max_y = (self.length / 2) * 1.25
            max_z = self.radius_outer

        spec = fig.add_gridspec(2, 2)

        side_view = fig.add_subplot(spec[0, 0], aspect="equal")
        top_view = fig.add_subplot(spec[0, 1], aspect="equal")
        section = fig.add_subplot(spec[1,:],aspect="equal")
        #section = fig.add_subplot(spec[1,:], projection='polar')

        
        top_y = max_y
        if self.groundlevel != None and self.groundlevel.habitat_axis_height < -max_y:
            top_y = -self.groundlevel.habitat_axis_height+0.5
        if self.earthworks != None:
            top_y = self.earthworks.radius_inner - self.earthworks.axis_height_from_ground + self.earthworks.depth_of_axis + 2

        side_view_max_x = max(top_y, max_x)

        side_view.set_xlim(-side_view_max_x, side_view_max_x)
        side_view.set_ylim(-max_y, top_y)
        side_view.title.set_text("Elevation section view")

        top_view.set_xlim(-max_x, max_x)
        top_view.set_ylim(-max_z, max_z)
        top_view.title.set_text("Plan section view")

        section.set_xlim(-0.2 * max_x, 0.2 * max_x)
        section.set_ylim(0.95*self._shells[0].radius_outer, max_x*0.82)
        section.title.set_text("Wall cross-section detail")

        # Martian atmosphere
        side_view.fill_between(x=[-side_view_max_x, side_view_max_x], 
            y1=[top_y, top_y],
            y2=[-max_y, -max_y],
            color="#FFBEA0")

        # Ground level
        if self.groundlevel != None:
            if self.groundlevel.habitat_axis_height == 1e3:
                side_view.fill_between(x=[-side_view_max_x, side_view_max_x], 
                y1=[-0.9*max_y, -0.9*max_y],
                y2=[-max_y, -max_y],
                color="#BE4628")
            elif self.earthworks != None:
                side_view.fill_between(x=[-side_view_max_x, side_view_max_x], 
                y1=[top_y-0.5, top_y-0.5],
                y2=[-max_y, -max_y],
                color="#BE4628")

            else:
                if self.orientation == "horizontal":
                    side_view.fill_between(x=[-side_view_max_x, side_view_max_x], 
                    y1=[-self.groundlevel.habitat_axis_height, -self.groundlevel.habitat_axis_height],
                    y2=[-max_y, -max_y],
                    color="#BE4628")
                elif self.orientation == "vertical":
                    side_view.fill_between(x=[-side_view_max_x, side_view_max_x], 
                    y1=[-self.length/2 - self.groundlevel.habitat_axis_height, -self.length/2 - self.groundlevel.habitat_axis_height],
                    y2=[-max_y, -max_y],
                    color="#BE4628")
        elif self.earthworks != None:
                side_view.fill_between(x=[-side_view_max_x, side_view_max_x], 
                y1=[top_y-2, top_y-2],
                y2=[-max_y, -max_y],
                color="#BE4628")

        top_view.set_facecolor("#BE4628")

        # Earthworks if present
        if self.earthworks != None:
            earthworks_axis = self.earthworks.radius_inner - self.earthworks.axis_height_from_ground
            earthworks_circ = plt.Circle([0, earthworks_axis], self.earthworks.radius_inner, color="#FFBEA0")
            side_view.add_patch(earthworks_circ)
        

        for shell_count, shell in enumerate(reversed(self._shells)):
            col = choice(["r", "b", "g", "c", "y"])
            if "steel" in shell.material.name:
                col = "#777777"
            elif "aluminium" in shell.material.name:
                col = "#999999"
            elif "STP Air" in shell.material.name:
                col = "#87CEEB"
            elif "plastic" in shell.material.name:
                col = "b"
            elif "foam" in shell.material.name:
                col = "#333333"
            elif "CO2" in shell.material.name:
                col = "#FFA073"
            
            # Small section
            circ = plt.Circle([0, 0], shell.radius_outer, color=col)

            # Side section
            rect = patches.Rectangle([-shell.radius_outer, -shell.length/2], 2*shell.radius_outer, shell.length, color=col)

            # Endcaps
            if self.endcap_type == "hemisphere":
                endcap_1 = plt.Circle([0, -shell.length/2], shell.radius_outer, color=col)
                endcap_2 = plt.Circle([0, shell.length/2], shell.radius_outer, color=col)

            elif self.endcap_type == "flat" and shell.radius_inner != 0:
                endcap_thickness = sum(shell.thickness for shell in self._shells[1:len(self._shells) - shell_count])
                endcap_1 = patches.Polygon([
                    [-shell.radius_outer, -shell.length/2], [-0.2*shell.radius_outer, -shell.length/2 - endcap_thickness], 
                    [0.2*shell.radius_outer, -shell.length/2 - endcap_thickness], [shell.radius_outer, -shell.length/2]],
                    color=col)

                endcap_2 = patches.Polygon([
                    [-shell.radius_outer, shell.length/2], [-0.2*shell.radius_outer, shell.length/2 + endcap_thickness], 
                    [0.2*shell.radius_outer, shell.length/2 + endcap_thickness], [shell.radius_outer, shell.length/2]],
                    color=col)

            if self.orientation == "horizontal":
                side_view.add_patch(circ)
                
                top_view.add_patch(endcap_1)
                top_view.add_patch(endcap_2)
                top_view.add_patch(rect)
            
            elif self.orientation == "vertical":
                top_view.add_patch(circ)

                side_view.add_patch(endcap_1)
                side_view.add_patch(endcap_2)
                side_view.add_patch(rect)

            # Wedge
            wedge = plt.Circle([0, 0], shell.radius_outer, color=col)
            section.add_patch(wedge)
        
    def create_static_shell(self, material, thickness, thermal_resistance=0, parallel_thermal_resistance=0):
        """
        Create a StaticShell that conforms around the outside of the outermost one
        """
        
        new_shell = StaticShell(self.orientation, material, self.radius_outer, 
        thickness, self.length, False, thermal_resistance, parallel_thermal_resistance)

        self.radius_outer += thickness
        self.verified = False
        self._shells.append(new_shell)
    
    def create_ground_level(self, habitat_axis_height=1e3, thermal_resistance=0):
        """
        Create a GroundLevel object

        Args:
            habitat_axis_height (float):        height of the habitat central axis above ground level, for horizontal orientation
                                                height of the lower end of the cylinder above ground level, for vertical orientation
                                                Defaults to a large value, indicating no thermal conduction contact (m)
            thermal_resistance (float):         thermal resistance between Habitat outer wall, default to 0 = no thermal contact (K/W)
        """
        assert self.groundlevel==None, "A GroundLevel has already been set"

        groundlevel = GroundLevel(habitat_axis_height, thermal_resistance)

        self.groundlevel = groundlevel

    def create_earthworks(self, radius_inner, depth_of_axis, axis_height_from_ground=None):
        """
        Create an Earthworks object for the habitat, to represent a tunnel or similar

        Args:
            radius_inner (float):               internal radius of the earthworks cylinder (m)
            depth_of_axis (float):              depth of the central axis of the earthworks cylinder below ground level (m)
            axis_height_from_ground (float):    height of the habitat axis from the ground (m). 
                                                Overwritten by a GroundLevel object, if it exists
        """
        if self.groundlevel != None:
            axis_height_from_ground = self.groundlevel.habitat_axis_height
        
        habitat_outer_radius = max([shell.radius_outer for shell in self._shells])

        if axis_height_from_ground + habitat_outer_radius > (2 * radius_inner):
            return(ValueError("Habitat collides with inside of Earthworks. Make the Earthworks bigger or lower the habitat"))
        
        if depth_of_axis < radius_inner:
            return(ValueError("Syrtis does not currently support Earthworks which breach the surface"))
        
        self.earthworks = Earthworks(radius_inner, depth_of_axis, axis_height_from_ground)

    def append_shell(self, shell):
        """
        Append a Shell of any type to the Habitat, checking for no overlaps
        """

        self.verified = False
        
        self._shells.append(shell)

        self.radius_outer = max(self.radius_outer, shell.radius_outer)
    
    def verify_geometry(self):
        """
        Checks the geometry of the habitat for errors
        """
        
        self._shells.sort()

        assert len(self._shells) >= 2, "Habitat requires at least two Shells, one enclosing the other"
        
        # Do any of the layers overlap, or do they have any gaps?
        
        overlap_error = False
        gap_error = False
        length_error = False
        for shell_count, shell in enumerate(self._shells):

            if shell_count != 0:
                # Check the shell inside this one to ensure a snug fit

                if round(self._shells[shell_count - 1].radius_outer, 8) > round(shell.radius_inner, 8):
                    # The Shell overlaps the inner edge of the current one
                    overlap_error = True
                    break

                if round(self._shells[shell_count - 1].radius_outer, 8) < round(shell.radius_inner, 8):
                    # The Shell overlaps the inner edge of the current one
                    gap_error = True
                    break
                    
                if round(self._shells[shell_count - 1].length, 8) > round(shell.length, 8):
                    length_error = True
                    break
            
            else:
                assert round(shell.radius_inner, 8) == 0, "There is no Shell at the centre of the Habitat. It must be present for a full calculation. Use a ConstrainedIdealGas Shell to simulate the pressurised space."
        
        assert not overlap_error, "Some Shells overlap each other"
        assert not gap_error, "Some Shells have gaps betweem them"
        assert not length_error, "Some Shells become shorter, not longer, moving away from the centre of the Habitat"
        

        # Is the outermost layer a solid and set to external?

        assert type(self._shells[-1].material) == Solid, "Outermost shell must be a Solid"

        for shell in self._shells:
            shell.external = False

        self._shells[-1].external = True

        # Set whole-Habitat radius_outer and length_outer
        self.radius_outer = self._shells[-1].radius_outer
        self.length_outer = self._shells[-1].length


        # If ground level is set, use it to calculate angle of contact
        if self.groundlevel != None and self.orientation == "horizontal":
            if self.groundlevel.habitat_axis_height > self.radius_outer:
                # Habitat is entirely above the ground
                self.ground_contact_angle = 0

            elif self.groundlevel.habitat_axis_height < (-self.radius_outer):
                # Habitat is entirely below ground
                self.ground_contact_angle = 180
            
            else:
                self.ground_contact_angle = np.rad2deg(np.arccos(
                    (self.radius_outer - self.groundlevel.habitat_axis_height) / self.radius_outer))

        
        self.exposed_convective_area()

        self.verified = True
    
    def exposed_convective_area(self):
        """
        Find the area exposed to convective losses, endcap and cylinder.
        Called during geometry verification
        """

        if self.groundlevel == None:
            self.exposed_area_cylinder = 2 * np.pi * self.radius_outer * self.length_outer

            if self.endcap_type == "hemisphere":
                self.exposed_area_endcap = 4 * np.pi * np.power(self.radius_outer, 2)
            
            elif self.endcap_type == "flat":
                self.exposed_area_endcap = 2 * np.pi * np.power(self.radius_outer, 2)
        
        elif self.groundlevel.habitat_axis_height > self.radius_outer:
            self.exposed_area_cylinder = 2 * np.pi * self.radius_outer * self.length_outer

            if self.endcap_type == "hemisphere":
                self.exposed_area_endcap = 4 * np.pi * np.power(self.radius_outer, 2)
            
            elif self.endcap_type == "flat":
                self.exposed_area_endcap = 2 * np.pi * np.power(self.radius_outer, 2)
        
        elif (self.groundlevel.habitat_axis_height < self.radius_outer and self.groundlevel.habitat_axis_height > -self.radius_outer and 
        self.orientation == "horizontal"):
            # Habitat is partially below ground level, so exposed area of cylinder and endcaps are reduced

            self.exposed_area_cylinder = 2 * np.pi * self.radius_outer * self.length_outer * 0.5 * (
                1 + np.cos(np.deg2rad(self.ground_contact_angle)))
            
            if self.endcap_type == "hemisphere":
                # Spherical cap area formula
                self.exposed_area_endcap = 2 * np.pi * self.radius_outer * (
                    self.radius_outer + self.groundlevel.habitat_axis_height) 
            
            elif self.endcap_type == "flat":
                # Segment formula area formula
                self.exposed_area_endcap = 2 * np.power(self.radius_outer, 2) * (np.pi - 
                    np.deg2rad(self.ground_contact_angle) + 0.5 * np.cos(np.deg2rad(2 * self.ground_contact_angle)))
        
        elif self.groundlevel.habitat_axis_height < -self.radius_outer and self.orientation == "horizontal":
            # The entire habitat is below ground level
            self.exposed_area_cylinder = 0
            self.exposed_area_endcap = 0
            
        elif (self.groundlevel.habitat_axis_height < self.radius_outer and self.groundlevel.habitat_axis_height > 0 and 
        self.orientation == "vertical"):
            # The cylinder of the habitat is above ground level but a hemsipherical endcap would be truncated
                self.exposed_area_cylinder = 2 * np.pi * self.radius_outer * self.length_outer

                if self.endcap_type == "hemisphere":
                    # One truncated cap, one full cap
                    # Spherical cap area formulae
                    self.exposed_area_endcap = 2 * np.pi * self.radius_outer * self.groundlevel.habitat_axis_height
                    self.exposed_area_endcap += 2 * np.pi * np.power(self.radius_outer, 2)
                
                elif self.endcap_type == "flat":
                    self.exposed_area_endcap = 2 * np.pi * np.power(self.radius_outer, 2)
        
        elif (self.groundlevel.habitat_axis_height < 0 and self.groundlevel.habitat_axis_height > -self.length_outer and
            self.orientation == "vertical"):
            # A section of the cylinder is below ground
            self.exposed_area_cylinder = 2 * np.pi * self.radius_outer * (
                    self.length_outer + self.groundlevel.habitat_axis_height)
                
            if self.endcap_type == "hemisphere":
                self.exposed_area_endcap = 2 * np.pi * np.power(self.radius_outer, 2)

            elif self.endcap_type == "flat":
                self.exposed_area_endcap = np.pi * np.power(self.radius_outer, 2)
        
        elif (self.groundlevel.habitat_axis_height < -self.length_outer and self.groundlevel.habitat_axis_height < -(self.length_outer + self.radius_outer) and
            self.orientation == "vertical"):
            # The whole cylinder is below ground, but a hemispherical endcap may protrude
            self.exposed_area_cylinder = 0

            if self.endcap_type == "hemisphere":
                self.exposed_area_endcap = 2 * np.pi * self.radius_outer * (
                    self.length_outer + self.radius_outer + self.groundlevel.habitat_axis_height)
            
            elif self.endcap_type == "flat":
                self.exposed_area_endcap = 0
        
        elif self.groundlevel.habitat_axis_height < -(self.length_outer + self.radius_outer) and self.orientation == "vertical":
            # Habitat is entirely underground
            self.exposed_area_cylinder = 0
            self.exposed_area_endcap = 0

    def direct_solar_area(self, solar_altitude, solar_azimuth):
        """
        Find the area of the habitat in direct sunlight
        Called by solar_gain_direct to find areas
        
        Args:
            solar_altitude (float): vertical angle of the sun above the horizon (degrees)
            solar_azimuth (float):  horizontal angle of the sun RELATIVE TO HABITAT AXIS (degrees) - 0=directly along axis
        """
        solar_altitude_rad = np.deg2rad(solar_altitude)
        solar_azimuth_rad = np.deg2rad(solar_azimuth)

        direct_solar_area = 0

        if self.groundlevel == None:
            if self.orientation == "horizontal":
                direct_solar_area += 2 * self.length_outer * self.radius_outer * (
                np.cos(solar_altitude_rad) * np.cos(solar_azimuth_rad) + np.sin(solar_azimuth_rad))
                
                if self.endcap_type == "flat":
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                        np.cos(solar_altitude_rad) * abs(np.cos(solar_azimuth_rad)))
                
                elif self.endcap_type == "hemisphere":
                    # Area of hemisphere projected onto the plane perpendicular to the Sun
                    # Two components are axial (plan view) and radial-ish (side view)
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                        np.cos(solar_altitude_rad) + 0.5 * np.sin(solar_altitude_rad)) * abs(np.cos(solar_azimuth_rad))

            elif self.orientation == "vertical":
                direct_solar_area += 2 * self.length_outer * self.radius_outer * (
                np.cos(solar_altitude_rad)) 

                if self.endcap_type == "flat":
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * np.sin(solar_altitude_rad) 
                
                elif self.endcap_type == "hemisphere":
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                        np.sin(solar_altitude_rad) + 0.5 * np.cos(solar_altitude_rad))
        
        elif self.groundlevel.habitat_axis_height > self.radius_outer:
            if self.orientation == "horizontal":
                direct_solar_area += 2 * self.length_outer * self.radius_outer * (
                np.cos(solar_altitude_rad) * np.cos(solar_azimuth_rad) + np.sin(solar_azimuth_rad))
                
                if self.endcap_type == "flat":
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                        np.cos(solar_altitude_rad) * abs(np.cos(solar_azimuth_rad)))
                
                elif self.endcap_type == "hemisphere":
                    # Area of hemisphere projected onto the plane perpendicular to the Sun
                    # Two components are axial (plan view) and radial-ish (side view)
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                        np.cos(solar_altitude_rad) + 0.5 * np.sin(solar_altitude_rad)) * abs(np.cos(solar_azimuth_rad))

            elif self.orientation == "vertical":
                direct_solar_area += 2 * self.length_outer * self.radius_outer * (
                np.cos(solar_altitude_rad)) 

                if self.endcap_type == "flat":
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * np.sin(solar_altitude_rad) 
                
                elif self.endcap_type == "hemisphere":
                    direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                        np.sin(solar_altitude_rad) + 0.5 * np.cos(solar_altitude_rad))
        
        elif (self.groundlevel.habitat_axis_height < self.radius_outer and self.groundlevel.habitat_axis_height > -self.radius_outer and 
        self.orientation == "horizontal"):
            # Habitat is partially below ground level, so exposed area of cylinder and endcaps are reduced
            direct_solar_area += 2 * self.length_outer * self.radius_outer * (
                np.cos(solar_altitude_rad) * np.cos(solar_azimuth_rad) * 0.5 * (1 + np.cos(np.deg2rad(self.ground_contact_angle)))
                + np.sin(solar_azimuth_rad)) 
            
            if self.endcap_type == "flat":
                direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                    np.cos(solar_altitude_rad) * abs(np.cos(solar_azimuth_rad))) * 0.5 * (
                1 + np.cos(np.deg2rad(self.ground_contact_angle)))
            
            elif self.endcap_type == "hemisphere":
                # Area of hemisphere projected onto the plane perpendicular to the Sun
                # Two components are axial (plan view) and radial-ish (side view)
                direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                    np.cos(solar_altitude_rad) + 0.5 * np.sin(solar_altitude_rad)) * abs(np.cos(solar_azimuth_rad)) * 0.5 * (
                1 + np.cos(np.deg2rad(self.ground_contact_angle)))
        
        elif self.groundlevel.habitat_axis_height < -self.radius_outer and self.orientation == "horizontal":
            # Habitat is entirely below ground level
            pass

        elif (self.groundlevel.habitat_axis_height < 0 and self.groundlevel.habitat_axis_height > -self.length_outer and
            self.orientation == "vertical"):
            direct_solar_area += 2 * (self.length_outer + self.groundlevel.habitat_axis_height) * self.radius_outer * (
                np.cos(solar_altitude_rad)) 

            if self.endcap_type == "flat":
                direct_solar_area += np.pi * np.power(self.radius_outer, 2) * np.sin(solar_altitude_rad) 
            
            elif self.endcap_type == "hemisphere":
                direct_solar_area += np.pi * np.power(self.radius_outer, 2) * (
                    np.sin(solar_altitude_rad) + 0.5 * np.cos(solar_altitude_rad))
        
        elif (self.groundlevel.habitat_axis_height < -self.length_outer and self.groundlevel.habitat_axis_height < -(self.length_outer + self.radius_outer) and
            self.orientation == "vertical"):
            # The whole cylinder is below ground, but a hemispherical endcap may protrude
            if self.endcap_type == "hemisphere":
                direct_solar_area += np.pi * np.power(self.radius_outer * 
                np.cos(np.arcsin((self.length_outer + self.groundlevel.habitat_axis_height)/self.radius_outer)), 2) * (
                    np.sin(solar_altitude_rad) + 0.5 * np.cos(solar_altitude_rad))
            
            elif self.endcap_type == "flat":
                pass
        
        elif self.groundlevel.habitat_axis_height < -(self.length_outer + self.radius_outer) and self.orientation == "vertical":
            pass
    
        return(direct_solar_area)
        
    def indirect_solar_area(self):
        """
        Find the area of the habitat indirect sunlight
        Called by solar_gain_direct and solar_gain_indirect to find areas

        Very similar to convective_area, except a vertically oriented Habitat with flat endcaps will not have the flat endcap counted
        """
        indirect_solar_area = self.exposed_area_cylinder + self.exposed_area_endcap

            
        if self.orientation == "horizontal":
            pass
        elif self.groundlevel == None:
            pass
        elif (self.groundlevel.habitat_axis_height < -self.length_outer and self.groundlevel.habitat_axis_height < -(self.length_outer + self.radius_outer) and
            self.orientation == "vertical"):
            # One flat endcap is above ground level but isn't visible to the surface of the ground: exclude it
            indirect_solar_area -= np.pi * np.power(self.radius_outer, 2)
        
        return(indirect_solar_area)

    def build_thermal_resistances(self, shell_temperatures, g):
        """
        Outputs the thermal resistances of each shell layer
        
        Args:
            shell_temperatures (list of floats):    list of len(self._shells)+1, referring to temperatures at the boundary
                                                    of each material. 0th is the inner wall temperature of innermost shell,
                                                    1st is outer wall of innermost/inner wall of second.
            g (float):                              gravity
        """
        assert len(shell_temperatures) == len(self._shells)+1, "length of 'shell_temperatures' must be equal to number of Shells plus one"

        shell_thermal_resistances = np.zeros((len(self._shells)))

        for shell_count, shell in enumerate(self._shells):
            T_avg = (shell_temperatures[shell_count] + shell_temperatures[shell_count+1]) / 2
            T_delta = shell_temperatures[shell_count+1] - shell_temperatures[shell_count]

            shell.calculate_thermal_resistance(T_avg, T_delta, g)

            shell_thermal_resistances[shell_count] = shell.thermal_resistance
        
        return(shell_thermal_resistances)

    def placeholder_convective_loss(self, T_wall, T_air):
        """
        Placeholder constant-h model for testing purposes
        """
        h = 1
        A = 2 * np.pi * self.length * self.radius_outer

        R_th = 1 / (h * A)

        Q = (T_wall - T_air) / R_th
        
        return(Q)
    
    def nusselt_sphere(self, air, v_air, T_air, T_wall, D):
        """
        Find the Nusselt number for a sphere in uniform flow
        Whitaker correlation, all properties evaulated at freestream temperature
        From Reference [2], Equation (7-36)

        Args:
            air (ConstrainedIdealGas):  object for the external flow
            v_air (float):              velocity of airflow (m/s)
            T_air (float):              temperature of the freestream (K)
            T_wall (float):             temperature of the wall (K)
            D (float):                  diameter (m)
        """
        Re = air.Re(T_air, D, v_air)
        Pr = air.Pr(T_air)
        mu_air = air.mu(T_air)
        mu_wall = air.mu(T_wall)

        if 0.7 >= Pr or Pr >= 380:
            print("Warning: heat transfer on hemispherical endcap correlation is out of validation range. Proceed with caution")
        
        if 3.5 >= Re or Re >= 80000:
            print("Warning: heat transfer on hemispherical endcap correlation is out of validation range. Proceed with caution")

        Nu_D = 2 + (
            (0.4 * np.power(Re, 0.4) + 0.06 * np.power(Re, 2/3)) * np.power(Pr, 0.4) * 
            np.power(mu_air / mu_wall, 0.25))
        
        return(Nu_D)
    
    def nusselt_plate_crossflow(self, air, v_air, T_air, T_wall, D):
        """
        Find the Nusselt number for crossflow over a plate
        Does the correct checking for laminar, turbulent etc
        
        Args:
            air (ConstrainedIdealGas):  object for the external flow
            v_air (float):              velocity of airflow (m/s)
            T_air (float):              temperature of the freestream (K)
            T_wall (float):             temperature of the wall (K)
            D (float):                  length scale (m)
        """
        transition_Re = 5e5

        T_film = (T_wall + T_air) / 2

        Re = air.Re(T_film, D, v_air)
        Pr = air.Pr(T_film)

        if Re < transition_Re:
            # Laminar correlation, Reference [2] Equation (7-21)
            Nu_D = 0.664 * np.power(Re, 0.5) * np.power(Pr, 1/3)
        
        elif Re > transition_Re and Re < 1e7:
            # Use partially-laminar correlation, Reference [2] Equation (7-24)
            Nu_D = (0.037 * np.power(Re, 0.8) - 871) * np.power(Pr, 1/3)
        
        else:
            print("Warning: heat transfer on flat plate correlation is out of validation range. Proceed with caution")
            Nu_D = (0.037 * np.power(Re, 0.8) - 871) * np.power(Pr, 1/3)


        return(Nu_D)

    def convective_loss_endcap_cross(self, air, v_air, T_air, T_wall):
        """
        Convective heat loss from the endcaps with uniform crossflow
        
        Args:
            air (ConstrainedIdealGas):  object for the external flow
            v_air (float):              velocity of airflow (m/s)
            T_air (float):              temperature of the freestream (K)
            T_wall (float):             temperature of the wall (K)
        """
        D = self._shells[-1].radius_outer * 2

        if self.endcap_type == "hemisphere":
            Nu_D = self.nusselt_sphere(air, v_air, T_air, T_wall, D)
            
        elif self.endcap_type == "flat":
            Nu_D = self.nusselt_plate_crossflow(air, v_air, T_air, T_wall, D)
            
        h = Nu_D * air.k((T_air + T_wall) / 2) / D

        Q_conv = h * self.exposed_area_endcap * (T_wall - T_air)

        return(Q_conv)
    
    def convective_loss_endcap_axial(self, air, v_air, T_air, T_wall):
        """
        Convective heat loss from the endcaps with uniform crossflow
        
        Args:
            air (ConstrainedIdealGas):  object for the external flow
            v_air (float):              velocity of airflow (m/s)
            T_air (float):              temperature of the freestream (K)
            T_wall (float):             temperature of the wall (K)
        """
        D = self._shells[-1].radius_outer * 2

        if self.endcap_type == "hemisphere":
            Nu_D = self.nusselt_sphere(air, v_air, T_air, T_wall, D)
        
        elif self.endcap_type == "flat":
            if self.orientation == "horizontal":
                T_film = (T_wall + T_air) / 2

                Re = air.Re(T_film, D, v_air)
                Pr = air.Pr(T_film)
                if 4000 >= Re or Re >= 15000:
                    print("Warning: heat transfer on flat endcap correlation is out of validation range. Proceed with caution")
                
                Nu_D = 0.228 * np.power(Re, 0.731) * np.power(Re, 1/3)
            
            elif self.orientation == "vertical":
                Nu_D = self.nusselt_plate_crossflow(v_air, T_air, T_wall, D)

        h = Nu_D * air.k((T_air + T_wall) / 2) / D

        Q_conv = h * self.exposed_area_endcap * (T_wall - T_air)

        return(Q_conv)

    def convective_loss_cylinder_cross(self, air, v_air, T_air, T_wall):
        """
        Convective heat loss from a cylinder with uniform crossflow.

        Args:
            air (ConstrainedIdealGas):  object for the external flow
            v_air (float):              velocity of airflow (m/s)
            T_air (float):              temperature of the freestream (K)
            T_wall (float):             temperature of the wall (K)
        """
        T_film = (T_wall + T_air) / 2

        D = self._shells[-1].radius_outer * 2

        Re = air.Re(T_film, D, v_air)
        Pr = air.Pr(T_film)

        # Churchill-Bernstein correlation, from Reference [1] Equation (7-35)

        Nu_D = 0.3 + (((0.62 * np.power(Re, 0.5) * np.power(Pr, 1/3.)) / (np.power(1 + np.power(0.4 / Pr, 2/3), 0.25))) *
        np.power(1 + np.power(Re / 282000, 0.625), 0.8))

        h = Nu_D * air.k(T_film) / D

        Q = h * self.exposed_area_cylinder * (T_wall - T_air)

        return(Q)

    def convective_loss_cylinder_axial(self, air, v_air, T_air, T_wall):
        """
        Convective heat loss from a cylinder with uniform axial flow.

        Args:
            air (ConstrainedIdealGas):  object for the external flow
            v_air (float):              velocity of airflow (m/s)
            T_air (float):              temperature of the freestream (K)
            T_wall (float):             temperature of the wall (K)
        """
        T_film = (T_wall + T_air) / 2

        L = self._shells[-1].length

        Nu_D = self.nusselt_plate_crossflow(air, v_air, T_air, T_wall, L)

        h = Nu_D * air.k(T_film) / L

        Q_conv = h * self.exposed_area_cylinder * (T_wall - T_air)

        return(Q_conv)
    
    def sky_temperature(self, T_air):
        """
        Calculate the sky temperature for radiative heat transfer
        Uses equation from References [2] and [3], validated against data from Reference [5]

        Args: 
            T_air:      air temperature (K)
        """

        T_sky = 0.0552 * np.power(T_air, 1.5)

        return(T_sky)

    def view_factor_ground(self):
        """
        Calculate the view factor to the habitat to flat ground
        Uses equations from Reference [4]
        """

        if self.orientation == "vertical" and self.earthworks == None:
            vf_cylinder = 0.5

            vf_endcap = 0.5

        elif self.orientation == "horizontal" and self.earthworks == None:
            vf_cylinder = 0.5 

            vf_endcap = 0.5
        
        elif self.earthworks != None:

            vf_cylinder = self.earthworks.view_factor_ground_cylinder(
                self.orientation, self.radius_outer, self.length_outer)
            
            if self.endcap_type == "hemisphere":
                vf_endcap = self.earthworks.view_factor_ground_hemisphere(
                    self.orientation, self.radius_outer, self.length_outer)
            
            elif self.endcap_type == "flat":
                vf_endcap = self.earthworks.view_factor_ground_disc(
                    self.orientation, self.radius_outer, self.length_outer)
            
        vf = ((vf_cylinder * self.exposed_area_cylinder) + 
            (vf_endcap * self.exposed_area_endcap)) / (self.exposed_area_cylinder + self.exposed_area_endcap)   

        return(vf)

    def radiative_loss_sky(self, T_wall):
        """
        Calculate radiative loss to the sky.

        Args:
            T_wall (float):             temperature of the wall (K)
        """
        
        vf_sky = 1 - self.view_factor_ground()

        Q_sky = 5.67e-8 * self._shells[-1].material.emit * np.power(T_wall, 4) * (
            self.exposed_area_cylinder + self.exposed_area_endcap) * vf_sky
        
        return(Q_sky)
    
    def radiative_gain_sky(self, T_air):
        """
        Calculate radiative loss to the sky.

        Args:
            T_air (float):              temperature of the ambient surrounding air (K)
        """

        T_sky = self.sky_temperature(T_air)
        
        vf_sky = 1 - self.view_factor_ground()
        
        Q_sky = 5.67e-8 * self._shells[-1].material.absorb * np.power(T_sky, 4) * (
            self.exposed_area_cylinder + self.exposed_area_endcap) * vf_sky

        # A negative sign is used for consistency with convention that +ve Q = heat loss
        return(-Q_sky)
    
    def radiative_loss_ground(self, T_wall):
        """
        Calculate radiative loss to the sky.

        Args:
            T_wall (float):             temperature of the wall (K)
        """

        vf_ground = self.view_factor_ground()

        Q_ground = 5.67e-8 * self._shells[-1].material.emit * np.power(T_wall, 4) * (
            self.exposed_area_cylinder + self.exposed_area_endcap) * vf_ground

        return(Q_ground)
    
    def radiative_gain_ground(self, T_ground):
        """
        Calculate radiative loss to the sky.

        Args:
            T_ground (float):              temperature of the ground(K)
        """

        vf_ground = self.view_factor_ground()

        Q_ground_gain = 5.67e-8 * self._shells[-1].material.absorb * np.power(T_ground, 4) * (
            self.exposed_area_cylinder + self.exposed_area_endcap) * vf_ground

        # A negative sign is used for consistency with convention that +ve Q = heat loss
        return(-Q_ground_gain)

    def solar_gain_direct(self, solar_altitude, solar_azimuth, solar_intensity):
        """
        Calculate heat gain to the outermost layer of the habitat, or into the centre if all outer layers have sufficient transparency
        
        Args:
            solar_altitude (float): vertical angle of the sun above the horizon (degrees)
            solar_azimuth (float):  horizontal angle of the sun RELATIVE TO HABITAT AXIS (degrees) - 0=directly along axis
            solar_intensity (float):power delivered by solar radiation after dust absorption, W/m2
        """

        if self.earthworks == None:
            solar_intensity_multipler = 1
        else:
            if self.orientation == "vertical":
                axis_height_from_ground = 0
            elif self.orientation == "horizontal":
                axis_height_from_ground = self.radius_outer

            if self.groundlevel != None:
                axis_height_from_ground += self.groundlevel.habitat_axis_height
            
            solar_intensity_multipler = self.earthworks.direct_solar_intensity_scaling(
                self.orientation, self.radius_outer, self.length_outer, axis_height_from_ground, 
                solar_altitude, solar_azimuth)

        direct_lit_area = self.direct_solar_area(solar_altitude, solar_azimuth)
        
        if self._shells[-1].material.transmit == 0:
            # Outer layer is opaque - all solar energy is delivered to outermost shell
            Q_solar_direct = solar_intensity * solar_intensity_multipler * direct_lit_area * self._shells[-1].material.absorb
        
        # A negative sign is used for consistency with convention that +ve Q = heat loss
        return(-abs(Q_solar_direct))
    
    def solar_gain_indirect(self, solar_altitude, solar_azimuth, solar_intensity, albedo_ground):
        """
        Calculate heat gain to the outermost layer of the habitat, or into the centre if all outer layers have sufficient transparency
        
        Args:
            solar_altitude (float): vertical angle of the sun above the horizon (degrees)
            solar_azimuth (float):  horizontal angle of the sun RELATIVE TO HABITAT AXIS (degrees) - 0=directly along axis
            solar_intensity (float):power delivered by solar radiation after dust absorption, W/m2
            albedo_ground (float):  albedo of the surface around the habitat
        """

        if self.earthworks == None:
            solar_intensity_multipler = 1
        else:
            if self.orientation == "vertical":
                axis_height_from_ground = 0
            elif self.orientation == "horizontal":
                axis_height_from_ground = self.radius_outer

            if self.groundlevel != None:
                axis_height_from_ground += self.groundlevel.habitat_axis_height
            
            solar_intensity_multipler = self.earthworks.indirect_solar_intensity_scaling(
                self.orientation, self.radius_outer, self.length_outer, axis_height_from_ground, 
                solar_altitude, solar_azimuth)

        indirect_lit_area = self.indirect_solar_area()
        
        if self._shells[-1].material.transmit == 0:
            # Outer layer is opaque - all solar energy is delivered to outermost shell
            Q_solar_indirect = solar_intensity * solar_intensity_multipler * albedo_ground * indirect_lit_area * self._shells[-1].material.absorb
        
        # A negative sign is used for consistency with convention that +ve Q = heat loss
        return(-Q_solar_indirect)

    def conductive_loss_fixed_resistance(self, T_wall, T_ground, R_ground):
        """
        Calculate thermal resistance losses for a fixed thermal resistance value

        Args:
            T_wall (float):     temperature of the outer wall (K)
            T_ground (float):   temperature of the far-field ground (K)
            R_ground (float):   thermal resistance from the outer wall to the ground (K/W)
        """

        Q_ground = (T_wall - T_ground) / R_ground
        
        return(Q_ground)

    def conductive_loss_horizontal_cylinder_steady(self, T_wall, T_ground, k_ground, R_wall):
        """
        Calculate the conductive loss from a horizontal cylinder in contact with the ground, steady-state solution
        
        Uses formulae from [1] for fully buried.
        Uses formulae from [9] and [10] for partially buried

        Args:
            T_wall (float):         temperature of the outer wall (K)
            T_ground (float):       temperature of the far-field ground (K)
            k_ground (float):       thermal conductivity of the far-field ground (W/m/K)
            R_wall (float):         thermal resistance of the entire wall, required for Biot number (K/W)
        """

        if self.groundlevel == None:
            return(0)
        if self.groundlevel.habitat_axis_height > self.radius_outer:
            # Habitat is entirely above the ground
            return (0)
        
        elif self.groundlevel.habitat_axis_height < (-self.radius_outer):
            # Habitat is entirely below the ground - use formula from [1] (and [9] if shallow buried)
            buried_depth = -self.groundlevel.habitat_axis_height
            depth_radius = buried_depth / self.radius_outer
            
            if depth_radius > 3:
                # Formula (1) from Table 3.5 in [1]
                S = 2 * np.pi * self.length_outer / np.log(4 * buried_depth / self.radius_outer)

                Q_ground = k_ground * S * (T_wall - T_ground)

            else:
                # Formulae 2-5 from [9]
                U_wall = 1 / (R_wall * 2 * np.pi * self.length_outer * self.radius_outer)
                Bi = U_wall * self.radius_outer / k_ground
                alpha_0 = np.log(depth_radius + np.sqrt(np.power(depth_radius, 2) - 1))

                U = Bi * (k_ground / self.radius_outer) / np.sqrt(
                    1 + np.power(Bi * alpha_0, 2) + (2 * Bi * alpha_0 / np.tanh(alpha_0)))
                
                Q_ground = U * (2 * np.pi * self.length_outer * self.radius_outer) * (T_wall - T_ground)

        
        else:
            # Habitat is partially buried
            # Use formulae from [9] and [10]

            buried_depth = -self.groundlevel.habitat_axis_height
            depth_radius = buried_depth / self.radius_outer

            U_wall = 1 / (R_wall * 2 * np.pi * self.length_outer * self.radius_outer)

            Bi = U_wall * self.radius_outer / k_ground

            theta_b = np.arccos(depth_radius)

            C1 = np.sqrt(1 - np.power(depth_radius, 2))
            C2 = depth_radius + (C1 / (Bi * theta_b))

            if C2 > 1:
                Nu = 2 / (theta_b * (np.pi - theta_b)) * C1 / np.sqrt(np.power(C2, 2) - 1) * (
                np.pi/2 - np.arctan(np.sqrt((C2 + 1) / (C2 - 1)) * np.tan(theta_b / 2)))
            else:# C2 <= 1:
                Nu = 1 / (theta_b * (np.pi - theta_b)) * C1 / np.sqrt(1 - np.power(C2, 2)) * np.log(
                (np.tan(theta_b / 2) + np.sqrt((1 - C2) / (1 + C2))) / 
                (np.tan(theta_b / 2) - np.sqrt((1 - C2) / (1 + C2))))
            
            U = Nu * k_ground / self.radius_outer

            Q_ground = U * (2 * np.pi * self.length_outer * self.radius_outer * (
                1 - (theta_b / np.pi))) * (T_wall - T_ground)
        
        return(Q_ground)

    def conductive_loss_vertical_cylinder_steady(self, T_wall, T_ground, k_ground):
        """
        Calculate the conductive loss from a vertical cylinder in contact with the ground, steady-state solution
        
        Uses formulae from [1] for partially buried, and for fully buried with a correction term

        Args:
            T_wall (float):         temperature of the outer wall (K)
            T_ground (float):       temperature of the far-field ground (K)
            k_ground (float):       thermal conductivity of the far-field ground (W/m/K)
        """

        if self.groundlevel == None:
            return(0)
        
        elif self.groundlevel.habitat_axis_height >= 0:
            return (0)

        elif self.groundlevel.habitat_axis_height < (-self.length_outer):
            # Habitat is partially below 
            buried_depth = -self.groundlevel.habitat_axis_height

            S = 2 * np.pi * buried_depth / np.log(2 * buried_depth / self.radius_outer)

            Q_ground = k_ground * S * (T_wall - T_ground)

        elif self.groundlevel.habitat_axis_height > (-self.length_outer):
            # Habitat is entirely below ground
            buried_depth = -self.groundlevel.habitat_axis_height

            S_spread = 2 * np.pi * buried_depth / np.log(2 * buried_depth / self.radius_outer)
            R_spread = 1 / (k_ground * S_spread)

            R_overburden = abs(buried_depth - self.length_outer) / (k_ground * self.radius_outer * 2)

            R = R_spread + R_overburden

            Q_ground = (T_wall - T_ground) / R

        
        return(Q_ground)

    def conductive_loss_hemisphere_steady(self, T_wall, T_ground, k_ground):
        """
        Calculate the conductive loss from a sphere, steady-state solution. Used for spherical endcaps
        
        Uses equation 3.1.7 from [7]

        Args:
            T_wall (float):         temperature of the outer wall (K)
            T_ground (float):       temperature of the far-field ground (K)
            k_ground (float):       thermal conductivity of the far-field ground (W/m/K)
        """
        if self.groundlevel == None:
            return(0)
        
        if self.orientation == "horizontal":
            if self.groundlevel.habitat_axis_height > self.radius_outer:
                return (0)
            else:
                buried_depth = -self.groundlevel.habitat_axis_height

                S = 4 * np.pi * self.radius_outer / (1 + 1 / (buried_depth/self.radius_outer + 1))

                Q_ground = k_ground * S * (T_wall - T_ground)
        
        elif self.orientation == "vertical":
            if self.groundlevel.habitat_axis_height > self.radius_outer:
                return(0)
            elif self.groundlevel.habitat_axis_height < -self.length_outer:
                # Bottom hemisphere only - assume the heat flux is half that of a sphere at equivalent depth
                buried_depth = -self.groundlevel.habitat_axis_height

                S = 2 * np.pi * self.radius_outer / (1 + 1 / (buried_depth/self.radius_outer + 1))

                Q_ground = k_ground * S * (T_wall - T_ground)

            elif self.groundlevel.habitat_axis_height > -self.length_outer:
                # Both hemisphere are partially buried
                # As a very janky approximation, assume this is equal to a sphere with centre at the mid-length
                equiv_buried_depth = -self.groundlevel.habitat_axis_height - (self.length_outer / 2)

                S = 4 * np.pi * self.radius_outer / (1 + 1 / (equiv_buried_depth/self.radius_outer + 1))

                Q_ground = k_ground * S * (T_wall - T_ground)


        return(Q_ground)

    def conductive_loss_disc_steady(self, T_wall, T_ground, k_ground):
        """
        Calculate the conductive loss from a disc in various orientations, steady-state solution. Used for flat endcaps
        
        Uses formula from [1]

        Args:
            T_wall (float):         temperature of the outer wall (K)
            T_ground (float):       temperature of the far-field ground (K)
            k_ground (float):       thermal conductivity of the far-field ground (W/m/K)
        """

        if self.groundlevel == None:
            return(0)
        
        elif self.orientation == "horizontal":
            if self.groundlevel.habitat_axis_height > self.radius_outer:
                # Habitat is entirely above the ground
                return (0)
            else:
                #print("Conduction loss from disc: flux patterns assumed to be the same as a sphere, with reduced area")
                buried_depth = -self.groundlevel.habitat_axis_height

                S = 2 * np.pi * self.radius_outer / (1 + 1 / (buried_depth/self.radius_outer + 1))

                Q_ground = k_ground * S * (T_wall - T_ground)
        
        elif self.orientation == "vertical":
            if self.groundlevel.habitat_axis_height > 0:
                # Habitat is entirely above the ground
                return (0)
            
            elif self.groundlevel.habitat_axis_height == 0:
                # Disc is resting on the surface
                S = 4 * self.radius_outer

                Q_ground = k_ground * S * (T_wall - T_ground)
            
            elif self.groundlevel.habitat_axis_height > -self.length:
                # Just the bottom disc 
                # Use Equation from Table 3-1 in [11], halved

                buried_depth = -self.groundlevel.habitat_axis_height

                S = 2 * np.pi * self.radius_outer / (
                    0.5 * np.pi - np.arctan(self.radius_outer / (2 * buried_depth)))
                
                Q_ground = k_ground * S * (T_wall - T_ground)
            
            else:
                # Both discs buried
                buried_depth_lower = -self.groundlevel.habitat_axis_height
                buried_depth_upper = -self.groundlevel.habitat_axis_height - self.length

                S_lower = 2 * np.pi * self.radius_outer / (
                    0.5 * np.pi - np.arctan(self.radius_outer / (2 * buried_depth_lower)))
                
                S_upper = 2 * np.pi * self.radius_outer / (
                    0.5 * np.pi - np.arctan(self.radius_outer / (2 * buried_depth_upper)))

                Q_ground = k_ground * (S_lower + S_upper) * (T_wall - T_ground)

        return(Q_ground) 



"""
Contains the Object which solves for a given Habitat and Configuration

"""

from syrtis.habitat import *
from syrtis.configuration import *


class Solver:
    def __init__(self, name, habitat, configuration):

        assert type(habitat) == Habitat, "'habitat' must be a Habitat object"
        assert type(configuration) == Configuration, "'configuration' must be a Configuration object"

        assert habitat.verified, "'habitat' must be verified with habitat.verify_geometry()"

        self.name = name
        self.habitat = habitat
        self.configuration = configuration

        self.heat = 0
        self.shell_temperatures = []
        self.report = {
                "Name":       self.name,
                "Total heat flux out": np.nan,
                "Outer wall temperature": np.nan,
                "Inner wall temperature": np.nan,
                "Convective loss from cylinder": 0,
                "Convective loss from endcap": 0,
                "Radiative loss to sky": 0,
                "Radiative loss to ground": 0,
                "Radiative gain from sky": 0,
                "Radiative gain from ground": 0,
                "Direct solar gain": 0,
                "Reflected solar gain": 0,
                "Conduction loss to ground": 0
            }
        self.thermal_energy = 0


    def solve(self):
        if self.configuration.solution_type == "constant temperature":
            self.iterate_constant_temperature()
        elif self.configuration.solution_type == "constant power":
            self.iterate_constant_power()

    def iterate_constant_temperature(self):
        """
        Iterate to find the correct properties for a constant inner wall temperature
        Requires a back-and-forth between conduction power loss and external power loss, which must be balanced
        Typically slower
        """
        T_internal_start = self.configuration.T_habitat
        Q_internal_flux = 1000
        
        Q_internal_flux = Q_internal_flux
        shell_temperatures, R_wall = self.generate_initial_state(T_internal_start)

        current_error = Q_internal_flux

        cutoff_ratio = 1e-5
        target_iterations = 1750

        iterations = 0

        while (abs(current_error) > cutoff_ratio) and iterations < target_iterations:

            new_shell_temperatures, R_wall = self.conduction_temperatures(shell_temperatures, Q_internal_flux)

            Q_external_flux = self.external_losses(new_shell_temperatures[-1], R_wall)

            new_error = abs(Q_internal_flux - Q_external_flux) / Q_internal_flux
            current_error = new_error
            iterations += 1

            Q_internal_flux = (Q_external_flux + (98 * Q_internal_flux)) / 99
            shell_temperatures = new_shell_temperatures
        
        if (abs(current_error) > cutoff_ratio):
            print("Did not converge " + str(abs(current_error)))
            return(np.nan)
        else:
            self.external_losses(shell_temperatures[-1], R_wall, verbose=True)

            self.shell_temperatures = shell_temperatures
            self.report["Inner wall temperature"] = shell_temperatures[0]
            self.heat = Q_external_flux
        
        """inputs = [*[Q_internal_flux], *shell_temperatures]

        minimised_outputs = minimize(self.minimisation, inputs, options={"ftol":cutoff_ratio, "xtol":cutoff_ratio})

        print(minimised_outputs)

        Q_internal_flux = minimised_outputs.x[0]
        shell_temperatures = minimised_outputs.x[1:]

        checked_shell_temperatures = self.conduction_temperatures(shell_temperatures, breakdown["Power loss"])
        print(checked_shell_temperatures)

        return(Q_internal_flux)"""
    
    def iterate_constant_power(self):
        """
        Iterate to find the shell temperatures when power loss from the habitat is constant
        """
        Q_target = self.configuration.Q_habitat
        T_internal_start = 273
        shell_temperatures, R_wall = self.generate_initial_state(T_internal_start)
        Q_loss = 0

        current_error = 1

        cutoff_ratio = 1e-5
        target_iterations = 2750

        iterations = 0

        while (abs(current_error) > cutoff_ratio) and iterations < target_iterations:
            Q_loss = self.external_losses(shell_temperatures[-1], R_wall)

            if Q_target != 0:
                current_error = abs((Q_target - Q_loss) / Q_target)
            else:
                current_error = abs(Q_loss) / 1000
            
            if current_error > 49:
                current_error = 49
            elif current_error < -49:
                current_error = -49

            if Q_loss < Q_target:
                # Actual and target power have the sign, but power loss is too low
                # External shell temperature needs to be increased
                shell_temperatures[-1] *= (1 + 0.02*current_error)
            elif Q_loss > Q_target:
                # Actual and target power have the sign, but power loss is too high
                # External shell temperature needs to be decreased
                shell_temperatures[-1] /= (1 + 0.02*current_error)
            
            new_shell_temperatures, R_wall = self.conduction_temperatures(shell_temperatures, Q_target)

            shell_temperatures = new_shell_temperatures

            #if iterations%10==0:print(shell_temperatures[0], shell_temperatures[-1], Q_loss, Q_target, current_error, R_wall)

            iterations += 1

        if (abs(current_error) > cutoff_ratio):
            print("Did not converge " + str(abs(current_error)), shell_temperatures, Q_loss)
            self.heat = np.nan
            return(np.nan)
        else:
            self.external_losses(shell_temperatures[-1], R_wall, verbose=True)
            
            self.report["Inner wall temperature"] = shell_temperatures[0]

            self.shell_temperatures = shell_temperatures
            self.heat = Q_target

    def verify_temperatures(self, shell_temperatures):
        """
        Verify that a given set of temperatures are either monotonically increasing or decreasing 
        This is a reasonable metric for a good solution, as it implies that flux is consistent through the Habitat
        May be violated in future if heating/cooling jackets are added as a Shell subclass

        Also verify that the external wall temperature is between the internal and environment temperatures
        """
        non_increasing = all(x>=y for x, y in zip(shell_temperatures, shell_temperatures[1:]))
        non_decreasing = all(x<=y for x, y in zip(shell_temperatures, shell_temperatures[1:]))

        temperature_correct = True
        
        if self.configuration.solution_type == "constant temperature":
            if self.configuration.T_habitat > self.configuration.T_air:
                # Habitat should be losing heat to the environment
                # External shell temperature should not be substantially hotter than the habitat
                if shell_temperatures[-1] > (self.configuration.T_habitat + 50):
                    temperature_correct = False
            
            elif self.configuration.T_habitat < self.configuration.T_air:
                # Habitat should be strictly gaining heat from the environment
                # External shell temperature should be greater than the habitat
                if shell_temperatures[-1] < self.configuration.T_habitat:
                    temperature_correct = False

        return((non_increasing or non_decreasing) and temperature_correct)

    def minimisation(self, inputs):
        """
        Minimise-able function, called into scipy.optimize solvers

        Args:
            inputs (list of floats):    list of all inputs to allow solving
                inputs[0]:              power loss from habitat as calculated by solve_wall_loss
                inputs[1:n]             shell temperatures for shells [0:n-1]
        """
        Q_internal_flux = inputs[0]
        shell_temperatures = inputs[1:]

        temperature_verification = self.verify_temperatures(shell_temperatures)
        if temperature_verification == False:
            return(1e99)

        else:
            new_shell_temperatures = self.conduction_temperatures(shell_temperatures, Q_internal_flux)

            Q_external_flux = self.external_losses(new_shell_temperatures[-1])

            error = abs(Q_internal_flux - Q_external_flux)**2

            return(error)

    def generate_initial_state(self, T_internal_start, T_external_start=0):
        # Assume there are (N+1) Shells, set the outermost - not attached to any physical Shell has temperature equal to outside air
        # The rest are linearly spaced up to the internal temperature
        if T_external_start == 0:
            T_external_start = self.configuration.T_air

        N_temperatures = len(self.habitat._shells) + 1

        T_external_wall = ((T_internal_start - T_external_start) / (N_temperatures + 1)) + T_external_start
        
        initial_shell_temperatures = np.linspace(T_internal_start, T_external_wall, N_temperatures)


        return(initial_shell_temperatures, 1)

    def conduction_temperatures(self, shell_temperatures, Q):
        """
        Find the updated temperatures across the habitat wall, based on a set of starting temperatures and the heat flux
        """

        wall_resistances = self.habitat.build_thermal_resistances(shell_temperatures, self.configuration.GRAVITY)

        total_resistance = sum(wall_resistances)

        updated_shell_temperatures = shell_temperatures

        # For both calculation types, the interior wall temperature is considered constant in this calculation
        # The temperatures throughout the rest of the Shells are found with the thermal resistances

        if self.configuration.solution_type == "constant temperature":
            # Iterate outwards from the inside to the outside - appropriate for a constant-temperature type solution
            updated_shell_temperatures[0] = self.configuration.T_habitat

            for shell_count in range(1,len(wall_resistances)+1):

                shell_temperature = updated_shell_temperatures[shell_count-1] - wall_resistances[shell_count-1] * Q
    
                if shell_temperature <= 0:
                    shell_temperature = shell_temperatures[shell_count]

                updated_shell_temperatures[shell_count] = shell_temperature
    
        elif self.configuration.solution_type == "constant power":
            # Assume the outermost shell temperature is correct and iterate inwards
            updated_shell_temperatures[-1] = shell_temperatures[-1]

            for shell_count in reversed(range(0, len(wall_resistances))):
                shell_temperature = updated_shell_temperatures[shell_count+1] + wall_resistances[shell_count] * Q

                if shell_temperature <= 0:
                    # This means power loss is too high for the wall geometry used
                    updated_shell_temperatures = [shell_temperatures[0] for i in shell_temperatures]
                    break
                
                updated_shell_temperatures[shell_count] = shell_temperature

        return(updated_shell_temperatures, total_resistance)
    
    def external_losses(self, wall_temperature, thermal_resistance_wall, verbose=False):

        #Q_wall = self.habitat.placeholder_convective_loss(wall_temperature, self.configuration.T_air)
        Q_wall = 0
        Q_endcap = 0
        Q_rad_sky_out = 0
        Q_rad_sky_in = 0
        Q_rad_ground_out = 0
        Q_rad_ground_in = 0
        Q_solar_direct = 0
        Q_solar_indirect = 0
        Q_conduction = 0

        convective = True
        solar = True
        radiative = True
        conductive = False

        if self.configuration.v_air == 0:
            convective = False

        if self.habitat.groundlevel != None:
            conductive = True
            if self.habitat.orientation == "horizontal" and self.habitat.groundlevel.habitat_axis_height < -self.habitat.radius_outer:
                convective = False
                solar = False
                radiative = False
            elif self.habitat.orientation == "vertical" and self.habitat.groundlevel.habitat_axis_height < -(
                self.habitat.length_outer + self.habitat.radius_outer):
                convective = False
                solar = False
                radiative = False

        if self.habitat.earthworks != None:
            if self.habitat.earthworks.axis_height_from_ground + self.habitat.radius_outer > (2 * self.habitat.earthworks.radius_inner):
                solar = False


        if convective:
            if (self.configuration.air_direction == "cross" and self.habitat.orientation == "horizontal") or (
                self.habitat.orientation == "vertical"):
                Q_wall = self.habitat.convective_loss_cylinder_cross(self.configuration.air, 
                                                                self.configuration.v_air,
                                                                self.configuration.T_air,
                                                                wall_temperature)
                
                Q_endcap = self.habitat.convective_loss_endcap_cross(self.configuration.air, 
                                                                self.configuration.v_air,
                                                                self.configuration.T_air,
                                                                wall_temperature)
            
            elif (self.configuration.air_direction == "axial" and self.habitat.orientation == "horizontal"):
                Q_wall = self.habitat.convective_loss_cylinder_axial(self.configuration.air, 
                                                                self.configuration.v_air,
                                                                self.configuration.T_air,
                                                                wall_temperature)
                
                Q_endcap = self.habitat.convective_loss_endcap_axial(self.configuration.air, 
                                                                self.configuration.v_air,
                                                                self.configuration.T_air,
                                                                wall_temperature)
        
        
        # Sign convention is preserved by functions in Habitat
        if radiative:
            Q_rad_sky_out = self.habitat.radiative_loss_sky(wall_temperature)
            Q_rad_sky_in = self.habitat.radiative_gain_sky(self.configuration.T_air)

            Q_rad_ground_out = self.habitat.radiative_loss_ground(wall_temperature)
            Q_rad_ground_in = self.habitat.radiative_gain_ground(self.configuration.T_ground)
        
        if solar:
            if self.configuration.solar_altitude > 0:
                Q_solar_direct = self.habitat.solar_gain_direct(self.configuration.solar_altitude,
                                                                self.configuration.solar_azimuth,
                                                                self.configuration.solar_intensity)
                
                Q_solar_indirect = self.habitat.solar_gain_indirect(self.configuration.solar_altitude,
                                                                self.configuration.solar_azimuth,
                                                                self.configuration.solar_intensity,
                                                                self.configuration.albedo_ground)
        if conductive:
            if self.habitat.groundlevel.thermal_resistance != 0:
                Q_conduction = self.habitat.conductive_loss_fixed_resistance(wall_temperature,
                                                                            self.configuration.T_ground,
                                                                            self.habitat.groundlevel.thermal_resistance)
            else:
                if self.habitat.orientation == "horizontal":
                    Q_conduction += self.habitat.conductive_loss_horizontal_cylinder_steady(wall_temperature,
                                                                                            self.configuration.T_ground,
                                                                                            self.configuration.k_ground,
                                                                                            thermal_resistance_wall)
                elif self.habitat.orientation == "vertical":
                    Q_conduction += self.habitat.conductive_loss_vertical_cylinder_steady(wall_temperature,
                                                                                        self.configuration.T_ground,
                                                                                        self.configuration.k_ground)
                
                if self.habitat.endcap_type == "hemisphere":
                    Q_conduction += self.habitat.conductive_loss_hemisphere_steady(wall_temperature,
                                                                                self.configuration.T_ground,
                                                                                self.configuration.k_ground)
                elif self.habitat.endcap_type == "flat":
                    Q_conduction += self.habitat.conductive_loss_disc_steady(wall_temperature,
                                                                            self.configuration.T_ground,
                                                                            self.configuration.k_ground)
        
        Q_convective = Q_wall + Q_endcap
        Q_rad_environment_out = Q_rad_sky_out + Q_rad_ground_out
        Q_rad_environment_in = Q_rad_sky_in + Q_rad_ground_in


        Q_total = (Q_convective + 
        Q_rad_environment_out + Q_rad_environment_in + 
        Q_solar_direct + Q_solar_indirect + 
        Q_conduction)
        
        if verbose:
            self.report["Total heat flux out"] = Q_total
            self.report["Outer wall temperature"] = wall_temperature
            self.report["Convective loss from cylinder"] = Q_wall
            self.report["Convective loss from endcap"] = Q_endcap
            self.report["Radiative loss to sky"] = Q_rad_sky_out
            self.report["Radiative loss to ground"] = Q_rad_ground_out
            self.report["Radiative gain from sky"] = Q_rad_sky_in
            self.report["Radiative gain from ground"] = Q_rad_ground_in
            self.report["Direct solar gain"] = Q_solar_direct
            self.report["Reflected solar gain"] = Q_solar_indirect
            self.report["Conduction loss to ground"] = Q_conduction
        else:
            return(Q_total)
    
    def calculate_thermal_energy(self, T_ref=0):
        """
        Calculates the total thermal energy stored in the habitat, compared to a reference temperature

        Args:
            T_ref (float):          reference temperature for the energy state (K)
        """
        assert len(self.shell_temperatures) == len(self.habitat._shells) + 1, "Run solver.solve before solver.thermal_energy"

        self.thermal_energy = 0

        for shell_count, shell in enumerate(self.habitat._shells):
            T_avg = sum(self.shell_temperatures[shell_count:shell_count+1])
            self.thermal_energy += shell.thermal_energy(self.habitat.endcap_type, T_avg, T_ref)

    def generate_error(self, state1, state2):
        error = np.sum((np.ndarray(state1) - np.ndarray(state2))**2)

        return(error)



"""
Allows installation via pip, e.g. by navigating to this directory with the command prompt, and using 'pip install .'
"""


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "syrtis",                     
    version = "0.1.0",                        
    author = "Sam Ross",                     
    author_email = "smross106@gmail.com",
    license = "MIT",
    description = "Thermal solver for crewed Martian habitats",
    keywords = ["habitat", "thermal", "mars", "heat"],
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),    # List of all python modules to be installed
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering"
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.6',               
    py_modules=["syrtis"],           
    install_requires=["numpy", "matplotlib"]                   
)


"""
Classes and inherited subclasses for Shells, the general term for layers of material included in the simulation.
In most cases, a Shell will be added as a subclass with more specialised properties

References:

 - [1] - https://en.wikipedia.org/wiki/Thermal_resistance
 - [2] - Y Cengel, Heat Transfer
 - [3] - Radiation View Factors, http://webserver.dmt.upm.es/~isidoro/tc3/Radiation%20View%20factors.pdf
 - [4] - Rapid Method for Determining Concentric Cylinder Radiation View Factors, Rea 1975

"""

from syrtis.material import *
import numpy as np

class Shell:
    """
    Object to store a single Shell. Parent to other subtypes of Shell, mostly empty
    """
    def __init__(self):
        pass

class StaticShell(Shell):
    """
    A Shell which forms a complete cylinderical layer around the centre of the object, without external flow.
    This can be a solid material (eg a layer of insulation or structural hull) or a layer of gas without forced flow.

    Args:
        orientation (str):        either 'horizontal' or 'vertical' to specify orientation of axis
        material (syrtis.Material): the material of the Shell
        radius_inner (float):       the internal radius of the Shell (m)
        thickness (float):          the thickness of the Shell (m)
        length (float):             the length of the Shell (m)
        external (bool) :           if True, the outer surface of the Shell is taken as the outer surface of the whole system. Default False
        thermal_resistance (float): an override for simple calculation of the Shell's thermal resistance. Ignored when not a positive number
    """

    def __init__(self, orientation, material, radius_inner, thickness, length, 
    external=False, thermal_resistance=0, parallel_thermal_resistance=0):

        assert orientation == "horizontal" or orientation == "vertical", "'orientation' must be either 'horizontal' or 'vertical'"

        assert type(material) in material_classes, "'material' input is not valid Material-inherited object"

        assert is_numeric(radius_inner, not_negative=True), "Shell 'internal_radius' must be a positive numerical value"

        assert is_numeric(thickness, positive=True), "Shell 'thickness' must be a positive numerical value"

        assert is_numeric(length, not_negative=True), "Shell 'length' must be a positive numerical value"

        self.orientation = orientation
        self.material = material

        self.radius_inner = radius_inner
        self.thickness = thickness
        self.radius_outer = self.radius_inner + self.thickness

        self.length = length

        self.external = external

        if thermal_resistance == 0:
            self.thermal_resistance_constant = 0
        else:
            self.thermal_resistance_constant = thermal_resistance
        
        self.parallel_thermal_resistance = parallel_thermal_resistance
        
    def __le__(self, other):
        if type(other) != StaticShell: raise TypeError

        return (self.radius_inner <= other.radius_inner)
    
    def __lt__(self, other):
        if type(other) != StaticShell: raise TypeError

        return (self.radius_inner < other.radius_inner)
    
    def __ge__(self, other):
        if type(other) != StaticShell: raise TypeError

        return (self.radius_inner >= other.radius_inner)
    
    def __gt__(self, other):
        if type(other) != StaticShell: raise TypeError

        return (self.radius_inner > other.radius_inner)
    
    def __eq__(self, other):
        if type(other) != StaticShell: raise TypeError

        return (self.radius_inner == other.radius_inner and self.radius_outer == other.radius_outer)

    def __ne__(self, other):
        if type(other) != StaticShell: raise TypeError

        return (self.radius_inner != other.radius_inner and self.radius_outer != other.radius_outer)

    def __repr__(self):
        return("A shell of {}, with an inner radius of {:.2f} and thickness of {:.3f} \n".format(
            self.material.name, self.radius_inner, self.thickness))

    def calculate_thermal_resistance(self, T_avg, T_delta, g):
        """
        Calculates the thermal resistance of the Shell
        Determines which equation to use and uses it
        """
        if self.thermal_resistance_constant != 0:
            thermal_resistance = self.thermal_resistance_constant
            
        elif self.radius_inner == 0:
            thermal_resistance = 0

        elif type(self.material) == Solid:
            thermal_resistance = self.thermal_resistance_solid()
        
        elif type(self.material) == ConstrainedIdealGas:
            thermal_resistance = self.thermal_resistance_annulus(T_avg, T_delta, g)
        
        if self.parallel_thermal_resistance == 0:
            self.thermal_resistance = thermal_resistance
        else:
            if thermal_resistance == 0:
                self.thermal_resistance = thermal_resistance
            else:
                self.thermal_resistance = 1 / ((1 / thermal_resistance) + (1 / self.parallel_thermal_resistance))
            
    def thermal_resistance_solid(self):
        # Using equation from Reference [1]
        R_th = np.log(self.radius_outer / self.radius_inner) / (2 * np.pi * self.material.k * self.length)

        return(R_th)

    def thermal_resistance_annulus(self, T_avg, T_delta, g):
        """
        Find the thermal resistance across a gas-filled annulus using natural convection correlations
        
        Args:
            T_avg (float):      Average temperature in the annulus (K)
            T_delta (float):    Temperature difference across the annulus (K)
            p (float):          Pressure in the annulus (Pa)
            g (float)           Gravitational acceleration (m/s2)
        """
        Ra = self.material.Ra(T_avg, T_delta, g, self.thickness)
        Pr = self.material.Pr(T_avg)
        k = self.material.k(T_avg)

        if self.orientation == "horizontal":
            # Using equation 9-56 from Reference [2] (via Raithby and Hollands, 1975)

            F_cyl = np.power(np.log(self.radius_outer/self.radius_inner), 4) / (
                np.power(self.thickness, 3) * np.power(
                    np.power(2 * self.radius_inner, -0.6) + np.power(2 * self.radius_outer, -0.6), 5))
            # Shape factor of the cylinders

            if F_cyl * Ra > 100:
                k_eq = (0.386 * np.power(Pr / (0.861 + Pr), 0.25) 
                * np.power(F_cyl * Ra, 0.25))
                k_eq = max(k_eq, 1)
            else:
                k_eq = 1
            
        elif self.orientation == "vertical":
            # Using equation 9-52 to 9-53 from Reference [2] from Berkovsky and Polevikov, 1977
            ratio = self.length / self.thickness

            if 1 <= ratio and 2 < ratio:
                k_eq = 0.18 * np.power((Pr * Ra) / (0.2 + Pr), 0.29)

                # Ensure model is working in the correct regime
                if (Ra * Pr) / (0.2 * Pr) < 1e3 and k_eq > 1:
                    print("Warning: heat transfer in vertical enclosure correlation is out of validation range. Proceed with caution")
                 
            elif 2 <= ratio < 10:
                k_eq = 0.22 * np.power((Pr * Ra) / (0.2 + Pr), 0.28) * np.power(ratio, -0.25)

                # Ensure model is working in the correct regime
                if Ra > 1e10 and k_eq > 1:
                    print("Warning: heat transfer in vertical enclosure correlation is out of validation range. Proceed with caution")
                
            elif 10 <= ratio:
                k_eq = 0.22 * np.power((Pr * Ra) / (0.2 + Pr), 0.28) * np.power(ratio, -0.25)

                if k_eq > 1:
                    print("Warning: heat transfer in vertical enclosure correlation is out of validation range. Proceed with caution")
            
            else:
                k_eq = 1
            
            if k_eq < 1:
                k_eq = 1

        R_th = np.log(self.radius_outer / self.radius_inner) / (2 * np.pi * k * k_eq * self.length)

        return(R_th)

    def thermal_energy(self, endcap_type, T_avg, T_ref=0):
        """
        Calculates the thermal energy (J) stored in the Shell relative to T_ref
        
        Args:
            endcap_type (str):      either "hemisphere" or "flat"
            T_avg (float):          average temperature of the Shell (K)
            T_ref (float):          reference temperature for the energy state (K)
        """

        volume = np.pi * (np.power(self.radius_outer, 2) - np.power(self.radius_inner, 2)) * self.length

        if endcap_type == "hemisphere":
            volume += (4/3) * np.pi * (np.power(self.radius_outer, 3) - np.power(self.radius_inner, 3))
        elif endcap_type == "flat":
            volume += 2 * np.pi * np.power(self.radius_outer, 2) * self.thickness
        
        if type(self.material) == Solid:
            thermal_capacity = self.material.rho * self.material.cp * volume
        elif type(self.material) == ConstrainedIdealGas:
            thermal_capacity = self.material.rho(T_avg) * self.material.cp(T_avg) * volume
        
        thermal_energy = thermal_capacity * (T_avg - T_ref)

        return(thermal_energy)

class GroundLevel(Shell):
    """
    An object to store the ground, whenever the Habitat is in contact with the ground

    Args:
        habitat_axis_height (float):        height of the habitat central axis above ground level, for horizontal orientation
                                            height of the lower end of the cylinder above ground level, for vertical orientation
                                            Defaults to a large value, indicating no thermal conduction contact (m)
        thermal_resistance (float):         thermal resistance between Habitat outer wall, default to 0 = no thermal contact (K/W)
    
    """
    def __init__(self, habitat_axis_height=1e3, thermal_resistance=0):
        
        assert is_numeric(habitat_axis_height), "GroundLevel 'height_below_habitat' must be a numerical value"

        assert is_numeric(thermal_resistance, not_negative=True), "GroundLevel 'thermal_resistance' must be a positive numerical value"

        self.habitat_axis_height = habitat_axis_height
        self.thermal_resistance = thermal_resistance

class Earthworks(Shell):
    """
    An object to store cylinderical hole in the ground in which the habitat sits. 
    Can include any large open cavity in the ground, including trenches or holes
    Also includes tunnels or lava tubes, natural or otherwise

    TODO: find a better name. Earthworks? Excavations? Cavities?

    Args:
        radius_inner (float):       radius of the inside of the of the cavity (m)
        depth_of_axis (float):      depth below ground level that the central axis of the cavity sits (m) 
                                    if less than radius_inner, the cavity will not be entirely below ground
        axis_height_from_ground (float):    height of the cylinder axis (horizontal) or bottom of cylinder (vertical)
                                    above the ground
    
    """
    def __init__(self, radius_inner, depth_of_axis, axis_height_from_ground):
        assert is_numeric(radius_inner, positive=True), "Cavity 'radius_inner' must be a positive number"
        assert is_numeric(depth_of_axis), "Cavity 'depth_of_axis' must be a number"
        assert depth_of_axis > -radius_inner, "Cavity 'depth_of_axis' is too low, cavity does not intersect the ground"

        self.radius_inner = radius_inner
        self.depth_of_axis = depth_of_axis
        self.axis_height_from_ground = axis_height_from_ground

        if self.depth_of_axis > self.radius_inner:
            self.exit_strip_width = 0
        else:
            self.exit_strip_width = 2 * self.radius_inner * np.sin(
                np.arccos(self.depth_of_axis / self.radius_inner))
    
    def view_factor_ground_cylinder(self, orientation, radius_outer, length_outer):
        """
        Calculates the view factor from a cylinder inside the Earthworks to the ground 
            (including the inner surface of the earthworks)
        
        Args:
            orientation (str):                  one of "horizontal" or "vertical"
            radius_outer (float):               outer radius of the cylinder (m)
            length_outer (float):               length of the cylinder

        """

        view_factor_ground = 0

        if self.depth_of_axis > self.radius_inner:
            # Earthworks has no view to the sky
            view_factor_ground = 1
        
        elif orientation == "horizontal":
            # Using Equation from [3] for view factor from planar strip to cylinder
            v = self.exit_strip_width / (2 * radius_outer)
            h = (self.radius_inner + self.depth_of_axis - self.axis_height_from_ground) / radius_outer

            if h > 0:
                # The top of the habitat is below the top rim of the hole
                # An virtual surface is drawn across the rim
                view_factor_ground = 1 - (np.arctan(v / h) / np.pi)
            
            else:
                # The top of the habitat extends above the top rim of the hole
                # Virtual surfaces are drawn above and to the sides of the cylinder

                v_top = 2 * radius_outer
                h_top = 1
                view_factor_sky_top = np.arctan(v_top / h_top) / np.pi

                # Side boxes
                w1 = self.radius_inner + self.depth_of_axis - self.axis_height_from_ground
                w2 = radius_outer
                h_side = 1

                v1_side = w1 / radius_outer
                v2_side = w2 / radius_outer

                view_factor_sky_side = (np.arctan(v2_side / h_side) - np.arctan(v1_side / h_side)) / np.pi

                view_factor_ground = 1 - (view_factor_sky_top + view_factor_sky_side)

        elif orientation == "vertical":

            exit_strip_radius = self.exit_strip_width / 2
            D = ((self.radius_inner + self.depth_of_axis) - (self.axis_height_from_ground + length_outer)) / exit_strip_radius
            Y = 1e10 / exit_strip_radius
            X = ((self.radius_inner + self.depth_of_axis) - self.axis_height_from_ground) / exit_strip_radius
            L = length_outer / exit_strip_radius
            R = radius_outer / exit_strip_radius

            def A(x):
                return(np.power(x, 2) + np.power(R, 2) - 1) 
            
            def B(x):
                return(np.power(x, 2) - np.power(R, 2) + 1)
            
            def F(x):
                f = np.sqrt(np.power((A(x) + 2) / R, 2 ) - 4) * np.arccos(A(x) * x / B(x)) - (
                    A(x) / (2 * x * R)) * np.arcsin(R)
                f *= (-0.5 / x) 
                f += B(x) / (8 * R * x)

                return(f)

            if self.radius_inner + self.depth_of_axis > (self.axis_height_from_ground + length_outer):
                # The top of the habitat is below the top rim of the hole
                # A virtual cylinder is drawn at the top of the rim
                # Equation 8 from [4] is used, with the length of the outer cylinder tending to infinity
                view_factor_sky = ((L + D) / D) * F(L + D)
                view_factor_sky += ((Y + D) / L) * F(Y + D)
                view_factor_sky -= (D / L) * F(D)
                view_factor_sky -= ((Y + D + L) / L) * F(Y + D + L)
            
            else:
                # The top of the habitat is above the top rim of the hole
                # A virtual cylinder is drawn at the top of the rim
                # Equation 6 from [4] is used, with the length of the outer cylinder tending to infinity

                view_factor_sky = (X / L) * F(X)
                view_factor_sky += ((L - X) / L) * (1 - F(L - X))
                view_factor_sky += ((Y + X - L) / L) * F(Y + X - L)
                view_factor_sky -= ((X + Y) / L) * F(X + Y)
            
            view_factor_ground = 1 - view_factor_sky

            
        return(view_factor_ground)

    def view_factor_ground_hemisphere(self, orientation, radius_outer, length_outer):
        """
        Calculates the view factor from a hemispherical endcap inside the Earthworks to the ground 
            (including the inner surface of the earthworks)
        
        Args:
            orientation (str):                  one of "horizontal" or "vertical"
            radius_outer (float):               outer radius of the cylinder (m)
            length_outer (float):               length of the cylinder
        """
        view_factor_ground = 0

        if self.depth_of_axis > self.radius_inner:
            # Earthworks has no view to the sky
            view_factor_ground = 1
        else:
            return(ValueError("Earthworks not currently configured for open-topped earthworks"))
        
        return(view_factor_ground)
    
    def view_factor_ground_disc(self, orientation, radius_outer, length_outer):
        """
        Calculates the view factor from a flat endcap inside the Earthworks to the ground 
            (including the inner surface of the earthworks)
        
        Args:
            orientation (str):                  one of "horizontal" or "vertical"
            radius_outer (float):               outer radius of the cylinder (m)
            length_outer (float):               length of the cylinder
        """
        view_factor_ground = 0

        if self.depth_of_axis > self.radius_inner:
            # Earthworks has no view to the sky
            view_factor_ground = 1
        else:
            return(ValueError("Earthworks not currently configured for open-topped earthworks"))
        
        return(view_factor_ground)
    
    def direct_solar_intensity_scaling(self, orientation, radius_outer, length_outer, solar_altitude, solar_azimuth):
        """
        Scales the direct solar intensity based on how much of the habitat can see the Sun at a given value of altitude and azimuth
        Passes value back to solar heat gain functions in Habitat 

        Args:
            orientation (str):                  one of "horizontal" or "vertical"
            radius_outer (float):               outer radius of the cylinder (m)
            length_outer (float):               length of the cylinder
            solar_altitude (float):             vertical angle of the sun above the horizon (degrees)
            solar_azimuth (float):              horizontal angle of the sun RELATIVE TO HABITAT AXIS (degrees) - 0=directly along axis
                                                assumed that a horizontal habitat lies in the same axis as the Earthworks tunnel
        
        TODO: work out the geometry for this function. The percentage of the habitat which is hit with direct sunlight.
            Probably requires working out the depth into the Earthworks which the sunlight reaches, and what fraction of the
            Habitat is above a given angle.
            Can't currently wrap head around how the solar azimuth affects the depth reached. When azimuth=0, sunlight will reach
            the bottom. When azimuth=90, it will be the direct projection from the rim. What about in between?
        """
        solar_multiplier = 1

        if self.depth_of_axis > self.radius_inner:
            # Earthworks has no view to the sky
            solar_multiplier = 0
            return(solar_multiplier)
        else:
            return(ValueError("Earthworks not currently configured for open-topped earthworks"))

        angle_over_rim = 0
        
        if orientation == "horizontal":
            pass
        
        elif orientation == "vertical":
            pass
        
        return(solar_multiplier)
    
    def indirect_solar_intensity_scaling(self, orientation, radius_outer, length_outer, solar_altitude, solar_azimuth):
        """
        Scales the indirect solar intensity based on how much of the habitat can see the Sun at a given value of altitude and azimuth
        Passes value back to solar heat gain functions in Habitat 

        Args:
            orientation (str):                  one of "horizontal" or "vertical"
            radius_outer (float):               outer radius of the cylinder (m)
            length_outer (float):               length of the cylinder
            solar_altitude (float):             vertical angle of the sun above the horizon (degrees)
            solar_azimuth (float):              horizontal angle of the sun RELATIVE TO HABITAT AXIS (degrees) - 0=directly along axis
                                                assumed that a horizontal habitat lies in the same axis as the Earthworks tunnel

        TODO: work out the geometry required for this function. The percentage of the surroundings lit with direct sunlight
        """
        solar_multiplier = 1

        if self.depth_of_axis > self.radius_inner:
            # Earthworks has no view to the sky
            solar_multiplier = 0
            return(solar_multiplier)
        else:
            return(ValueError("Earthworks not currently configured for open-topped earthworks"))
        
        if orientation == "horizontal":
            pass
        
        elif orientation == "vertical":
            pass
        
        return(solar_multiplier)
